/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

/*! \brief Add/remove entries from multiple threads.

This example shows that other mechanisms than tree locks may be employed to
prevent concurrent access to a tree from multiple threads.

This example builds on c_offset_fixed_int_tree.c by using the same tree but with
producer and consumer threads.

The "producer" thread supplies entries.
Two "consumer" threads remove entries.

Rather than have threads "spinning" pthread conditions and mutexes are used.

Execution proceeds as follows:
 -# Producer produces 2 entries and waits for the count to drop below 2.
 -# Consumers wait on available entries and are activitated once the producer
    notifies. They compete to consume the entries in the tree.
 -# Once the producer has produced 10 items it sets nDone to 1 and exits.
 -# The consumers empty the tree and see that it's time to quit and exit.
*/

#ifdef _PTHREAD_H
#error _PTHREAD_H defined.
#endif

#include "andromeda-avl-offset-int.h"

#include <pthread.h>
#include <error.h>
#include <v3c/time.hpp>
#include <v3c/utils.h>

pthread_barrier_t barrier;
volatile int nAvail = 0, nDone = 0;
pthread_mutex_t mtx;
pthread_cond_t cond;
AvlOffsetInt_runtime_t * runtime;

const char * sz_consumer[] =
{ "Consumer 1"
, "Consumer 2"
};
unsigned int count[2] = {0,0};
const char * sz_producer = "Producer";
const char * sz_main = "Main";

void check_return(const char * from, const char * msg, int err)
{
	if(!err)
		return;
	error(1, err, "%s : %s", from, msg);
}

void * producer(void *p)
{
	(void)p;
	printf("Producer: entered.\n");
	unsigned int n;

	int ret = pthread_barrier_wait(& barrier);
	if(ret != -1)
		check_return(sz_producer, "pthread_barrier_wait() failed", ret);
	printf("Producer passed the barrier.\n");

	printf("Producer: locking mutex.\n");
	ret = pthread_mutex_lock(& mtx);
	check_return(sz_producer, "pthread_mutex_lock() failed", ret);

	for(n = 0; n < 10u; n += 2) {
		if(n) {
			printf("Producer: calling pthread_cond_wait().\n");
			ret = pthread_cond_wait(& cond, & mtx); /* Unlock the mutex until a notify.*/
			check_return(sz_producer, "pthread_cond_broadcast() failed", ret);
		}
		if(nAvail >= 2) { /*?*/
			n -= 2;
			//pthread_cond_signal(& cond);
			continue;
		}
		printf("%s: inserting %u and %u.\n", sz_producer, n, n+1);
		runtime_insert(runtime, n);
		runtime_insert(runtime, n+1);
		nAvail += 2;
		printf("Producer: calling pthread_cond_broadcast().\n");
		ret = pthread_cond_broadcast(& cond);
		check_return(sz_producer, "pthread_cond_broadcast() failed", ret);
		milli_sleep(10);
	}
	printf("Producer: unlocking mutex.\n");
	ret = pthread_mutex_unlock(& mtx);
	check_return(sz_producer, "pthread_mutex_unlock() failed", ret);

	printf("Producer: exit.\n");
	nDone = 1;
	return 0;
}

void * consumer(void * p)
{
	unsigned int u = (unsigned int)((char *)p - (char *)0);

	printf("%s: entry.\n", sz_consumer[u]);
	int ret = pthread_barrier_wait(& barrier);
	if(ret != -1)
		check_return(sz_consumer[u], "pthread_barrier_wait() failed", ret);
	printf("%s: passed the barrier.\n", sz_consumer[u]);

	printf("%s: locking mutex.\n", sz_consumer[u]);
	ret = pthread_mutex_lock(& mtx);
	check_return(sz_consumer[u], "pthread_mutex_lock() failed", ret);

	int n = 0;
	for(;!nDone || nAvail; n = 1) {
		milli_sleep(10);
		if(n) {
			printf("%s: calling pthread_cond_wait().\n", sz_consumer[u]);
			ret = pthread_cond_wait(& cond, & mtx); /* Unlock the mutex until a notify.*/
			check_return(sz_consumer[u], "pthread_cond_wait() failed", ret);
		}
		switch(nAvail) {
		case -1:
			break;
		case 0:
			if(nDone) {
				printf("%s: finished.\n", sz_consumer[u]);
				break;
			}
			/* We were woken up but have nothing to do.
			   Give another thread a try.
			*/
			printf("%s: nothing to do - calling pthread_cond_signal().\n"
				, sz_consumer[u]);
			ret = pthread_cond_signal(& cond);
			check_return(sz_consumer[u], "pthread_cond_signal() failed", ret);
			continue;
		default: {
			int16_t item = -1;
			AvlOffsetInt_node_ptr_t pn = 0;
			(void)AvlOffsetInt_search_closest(runtime, runtime->m_pTree, 0
				, & pn);
			if(pn) {
				item = pn->m_item;
				printf("%s: removing %d.\n", sz_consumer[u], item);
				runtime_remove(runtime, item);
				--nAvail;
				printf("%s: calling pthread_cond_signal().\n", sz_consumer[u]);
				ret = pthread_cond_signal(& cond);
				check_return(sz_consumer[u], "pthread_cond_signal() failed"
					, ret);
				++count[u];
			}
			break;
			}
		}
	}
	printf("%s: unlocking mutex.\n", sz_consumer[u]);
	ret = pthread_mutex_unlock(& mtx);
	check_return(sz_consumer[u], "pthread_mutex_unlock() failed", ret);

	printf("%s: consumed %u entries : exit.\n", sz_consumer[u], count[u]);
	return 0;
}

#define START_OFFS V3C_ROUND_UP_T2(uint16_t, sizeof(AvlOffsetInt_tree_t), \
	TREEDB_ALIGN)

int main(int argc, char *argv[])
{
	(void)argc;
	(void)argv;
	printf("%s: entry.\n", sz_main);

	AvlOffsetInt_int_storage_t storage;
#ifndef __cplusplus
	storage.m_pVtbl = & AvlOffsetInt_storage_vtbl;
#endif
	char mem[
		START_OFFS + TREEDB_PAGE_SIZE
	] __attribute__ ((aligned (TREEDB_ALIGN))) = {0};
	// We need a base that can't be used - it's our NULL reference.
	// We also reserve the tree root at the start.
	AvlOffsetInt_container_t container
		= { mem - TREEDB_ALIGN, mem + START_OFFS };
	AvlOffsetInt_tree_t * ptree = (AvlOffsetInt_tree_t *)mem;

	AvlOffsetInt_runtime_t han =
	{ (AvlOffsetInt_storage_ptr_t)(& storage)
	, & container
	, ptree
	, PTHREAD_RWLOCK_INITIALIZER };

	runtime = & han;

	AvlOffsetInt_init(& han, han.m_pTree);

	printf("%s: tree created, calling pthread_barrier_init().\n", sz_main);
	int ret = pthread_barrier_init(& barrier, 0, 4);
	check_return(sz_main, "pthread_barrier_init() failed", ret);

	printf("%s: calling pthread_cond_init().\n", sz_main);
	ret = pthread_cond_init(& cond, 0);
	check_return(sz_main, "pthread_cond_init() failed", ret);

	printf("%s: calling pthread_mutex_init().\n", sz_main);
	ret = pthread_mutex_init(&mtx, 0);
	check_return(sz_main, "pthread_mutex_init() failed", ret);

	printf("%s: creating threads.\n", sz_main);
	pthread_t tp, tc1, tc2;
	ret = pthread_create(&tp, 0, producer, &han);
	check_return(sz_main, "pthread_create(producer) failed", ret);
	ret = pthread_create(&tc1, 0, consumer, (void *)0);
	check_return(sz_main, "pthread_create(consumer 1) failed", ret);
	ret = pthread_create(&tc2, 0, consumer, (void *)1);
	check_return(sz_main, "pthread_create(consumer 2) failed", ret);

	printf("%s: calling pthread_barrier_wait().\n", sz_main);
	ret = pthread_barrier_wait(& barrier);
	if(ret != -1)
		check_return(sz_main, "pthread_barrier_wait() failed", ret);
	printf("%s: passed the barrier.\n", sz_main);

	ret = pthread_join(tp, 0);
	check_return(sz_main, "pthread_join(producer) failed", ret);
	ret = pthread_join(tc1, 0);
	check_return(sz_main, "pthread_join(consumer 1) failed", ret);
	ret = pthread_join(tc2, 0);
	check_return(sz_main, "pthread_join(consumer 2) failed", ret);
	
	runtime_clear(& han);

	ret = pthread_cond_destroy(& cond);
	check_return(sz_main, "pthread_cond_destroy() failed", ret);
	ret = pthread_mutex_destroy(& mtx);
	check_return(sz_main, "pthread_mutex_destroy() failed", ret);
	ret = pthread_barrier_destroy(& barrier);
	check_return(sz_main, "pthread_barrier_destroy() failed", ret);

	printf("%s: consumers consumed %u / 10 items.\n", sz_main, count[0] + count[1]);
	return (count[0] + count[1]) == 10 ? 0 : 1;
}
