/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

/*! \brief Add/remove entries from multiple threads.

This example shows that other mechanisms than tree locks may be employed to
prevent concurrent access to a tree from multiple threads.

This example builds on c_offset_fixed_int_tree.c by using the same tree but with
producer and consumer threads.

The "producer" thread supplies entries.
Two "consumer" threads remove entries.

Rather than have threads "spinning" pthread conditions and mutexes are used.

Execution proceeds as follows:
 -# Producer produces 2 entries and waits for the count to drop below 2.
 -# Consumers wait on available entries and are activitated once the producer
    notifies. They compete to consume the entries in the tree.
 -# Once the producer has produced 10 items it sets nDone to 1 and exits.
 -# The consumers empty the tree and see that it's time to quit and exit.
*/

#include "andromeda-avl-offset-int.h"
#include <error.h>
#include <v3c/time.hpp>
#include <v3c/utils.h>

volatile int nDone = 0;
AvlOffsetInt_runtime_t * runtime;

const char * sz_consumer[] =
{ "Consumer 1"
, "Consumer 2"
};
unsigned int count[2] = {0,0};
const char * sz_producer = "Producer";
const char * sz_main = "Main";

void check_return(const char * from, const char * msg, int err)
{
	if(!err)
		return;
	error(1, err, "%s : %s", from, msg);
}

void * producer(void *p)
{
	(void) p;
	printf("Producer: entered.\n");
	unsigned int n;

	for(n = 0; n < 20u; n += 2) {
		runtime_write_lock(runtime);
		printf("%s: inserting %u and %u.\n", sz_producer, n, n+1);
		runtime_insert(runtime, n);
		runtime_insert(runtime, n+1);
		runtime_unlock(runtime);
		milli_sleep(10);
	}
	printf("Producer: exit.\n");
	nDone = 1;
	return 0;
}

void * consumer(void * p)
{
	unsigned int u = (unsigned int)((char *)p - (char *)0);

	printf("%s: entry.\n", sz_consumer[u]);
	int got_item = 0;
	for(;got_item || !nDone;) {
		got_item = 0;
		runtime_write_lock(runtime);
		AvlOffsetInt_node_ptr_t pNode = runtime_top(runtime);
		if(!pNode) {
			runtime_unlock(runtime);
			continue;
		}
		unsigned int nAvail = pNode->m_count;
		switch(nAvail) {
		case -1:
			break;
		case 0:
			if(nDone) {
				printf("%s: finished.\n", sz_consumer[u]);
				break;
			}
			/* We were woken up but have nothing to do.
			   Give another thread a try.
			*/
			printf("%s: nothing to do.\n", sz_consumer[u]);
			runtime_unlock(runtime);
			continue;
		default: {
			AvlOffsetInt_node_ptr_t pn = 0;
			(void)AvlOffsetInt_search_closest(runtime, runtime->m_pTree, 0
				, & pn);
			if(pn) {
				printf("%s: removing %d.\n", sz_consumer[u], pn->m_item);
				runtime_remove(runtime, pn->m_item);
				++count[u];
				got_item = 1;
			}
			break;
			}
		}
		runtime_unlock(runtime);
		milli_sleep(10);
	}
	printf("%s: consumed %u entries : exit.\n", sz_consumer[u], count[u]);
	return 0;
}

#define START_OFFS V3C_ROUND_UP_T2(uint16_t, sizeof(AvlOffsetInt_tree_t), \
	TREEDB_ALIGN)

int main(int argc, char *argv[])
{
	(void)argc;
	(void)argv;
	printf("%s: entry.\n", sz_main);

	AvlOffsetInt_int_storage_t storage;
#ifndef __cplusplus
	storage.m_pVtbl = & AvlOffsetInt_storage_vtbl;
#endif
	char mem[
		START_OFFS + TREEDB_PAGE_SIZE
	] __attribute__ ((aligned (TREEDB_ALIGN))) = {0};
	// We need a base that can't be used - it's our NULL reference.
	// We also reserve the tree root at the start.
	AvlOffsetInt_container_t container
		= { mem - TREEDB_ALIGN, mem + START_OFFS };
	AvlOffsetInt_tree_t * ptree = (AvlOffsetInt_tree_t *)mem;

	AvlOffsetInt_runtime_t han =
	{ (AvlOffsetInt_storage_ptr_t)(& storage)
	, & container
	, ptree
	, PTHREAD_RWLOCK_INITIALIZER };

	runtime = & han;

	AvlOffsetInt_init(& han, han.m_pTree);

	printf("%s: tree created, locking the tree.\n", sz_main);
	runtime_write_lock(runtime);

	printf("%s: creating threads.\n", sz_main);
	pthread_t tp, tc1, tc2;
	int ret = pthread_create(&tp, 0, producer, &han);
	check_return(sz_main, "pthread_create(producer) failed", ret);
	ret = pthread_create(&tc1, 0, consumer, (void *)0);
	check_return(sz_main, "pthread_create(consumer 1) failed", ret);
	ret = pthread_create(&tc2, 0, consumer, (void *)1);
	check_return(sz_main, "pthread_create(consumer 2) failed", ret);

	printf("%s: threads created, unlocking the tree.\n", sz_main);
	runtime_unlock(runtime);

	ret = pthread_join(tp, 0);
	check_return(sz_main, "pthread_join(producer) failed", ret);
	ret = pthread_join(tc1, 0);
	check_return(sz_main, "pthread_join(consumer 1) failed", ret);
	ret = pthread_join(tc2, 0);
	check_return(sz_main, "pthread_join(consumer 2) failed", ret);
	
	runtime_clear(& han);

	printf("%s: consumers consumed %u / 20 items.\n", sz_main, count[0] + count[1]);
	return (count[0] + count[1]) == 20 ? 0 : 1;
}
