/* Copyright (C) 2011  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#include <malloc.h>
#include <stdio.h>
#include <mcheck.h>

#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>

#include <sys/types.h> /* umask */
#include <sys/stat.h> /* umask */
#include <ctype.h> /* isprint */
#include <sys/wait.h> // signal

#define KIBIBYTE ((size_t)1024)
#define MEBIBYTE (KIBIBYTE * KIBIBYTE)
#define GIBIBYTE (KIBIBYTE * MEBIBYTE)

#include <dlfcn.h>
#include <v3c/smart_ptr.hpp>
#include <v3c/trace_scope.hpp>
#include <v3c/filesystem.hpp>

using namespace v3c;
using namespace std;

void sigsegv_handler(int signum)
{
	fprintf(stderr, "segfault\n");
	signal(signum, SIG_DFL);
	raise(signum);
}

typedef int (* DIDALLOC_FN)(void */*__ptr*/);

DIDALLOC_FN didalloc_fn = 0;

int default_didalloc(void *__ptr)
{
	(void)__ptr;
	return 1; // Assume yes.
}

void my_free(void *__ptr)
{
	if((*didalloc_fn)(__ptr)) {
		free(__ptr);
		return;
	}
	abort();
}

int using_treedb = 0;

int main(int argc, char * argv[])
{
	void * h = dlopen(0, RTLD_NOW);
	didalloc_fn = (DIDALLOC_FN)dlsym(h, "didalloc");
	dlclose(h);
	if(!didalloc_fn) {
		didalloc_fn = default_didalloc;
		fprintf(stderr, "Using regular malloc\n");
	}else{
		fprintf(stderr, "Using treedb malloc\n");
		using_treedb = 1;
	}
	CoutTraceWriter ctw("");

	smart_ptr< FileTraceWriter > ftw;
	if(using_treedb && v3c::fs::file_type("fuse/trace")) {
		ftw = new FileTraceWriter();
		ftw->open("fuse/trace", ios_base::out | ios_base::app);
	}
	signal(SIGSEGV, & sigsegv_handler);

	mcheck(0);
	TraceWrap("main");

	if(0) {
		int n;
		for(n = 0; n < argc; ++n)
			fprintf(stderr, ", \"%s\"", argv[n] ? argv[n] : "<null>");
		fprintf(stderr, "\n");
	}
	size_t sz;
	if(argc < 2)
		sz = 1; /* not a stress test */
	else{
		int len = strlen(argv[1]);
		errno = 0;
		char * tailptr;
		sz = strtoull(argv[1], & tailptr, 10);
		if(errno) {
			fprintf(stderr, "Error : %s isn't a number.\n", argv[1]);
			return 1;
		}
		int diff = tailptr - argv[1];
		size_t mult = 1;
		if(diff == len)
			; /* The whole number was parsed */
		else if(diff == (len - 1)) {
			switch(*tailptr) {
			case 'K':
				mult = KIBIBYTE;
				break;
			case 'M':
				mult = MEBIBYTE;
				break;
			case 'G':
				mult = GIBIBYTE;
				break;
			default:
				fprintf(stderr, "Error : %s isn't a number.\n", argv[1]);
				return 1;
			}
			if(sz * mult < sz) { /* overflow */
				fprintf(stderr, "Error : %s is too big.\n", argv[1]);
				return 1;
			}
			sz *= mult;
		}else{
			fprintf(stderr, "Error : %s isn't a number.\n", argv[1]);
			return 1;
		}
	}
	mcheck(0);
	trace << getpid() << ": will malloc " << sz << " bytes." << endl;
	mcheck(0);
	char * p = (char *)malloc(sz);
	if(!p) {
		trace_cerr << "malloc failed." << endl;
		return 1;
	}
	trace << "Calling fork()." << endl;
	pid_t pid = fork();
	if(pid == -1) {
		trace_cerr << "fork failed." << endl;
		exit(1);
	}
	mcheck(0);
	int res = 0, match_res = 0;
	size_t n;
	if(pid) { /* parent */
		TraceWrap("parent");
		for(n = 0; n < sz; n += 4096)
			p[n] = 'A';
		for(n = 0; n < sz; n += 4096)
			if(p[n] != 'A') {
				char c = '.';
				if(isprint(p[n]))
					c = p[n];
				trace_cerr << "Parent: comparison failed at " << n << " : "
					"wanted 'A'(" << (int)'A' << "), got '" << c << "'("
					<< (int)c << ")" << endl;
				match_res = 1;
				break;
			}
		int status;
		waitpid(pid, & status, 0);
		res = 0;
		if(!WIFEXITED(status)) {
			trace_cerr << "Parent: child exited abnormally." << endl;
			res = 1;
		}else{
			trace_cerr << "Parent: child exited normally with status "
				<< WEXITSTATUS(status) << "." << endl;
		}
		if(WIFSIGNALED(status)) {
			trace_cerr << "Parent: child exited with signal "
				<< strsignal(WTERMSIG(status)) << "(" << WTERMSIG(status)
				<< ")" << endl;
			res = 1;
		}
		res += match_res;
		my_free(p);
		trace << "Parent: calling mcheck()." << endl;
		mcheck(0);
		if(res)
			trace_cerr << "Parent: Failed." << endl;
		else
			trace << "Parent: Success." << endl;
	}else{ /* child */
		TraceWrap("child");
		for(n = 0; n < sz; n += 4096)
			p[n] = 'B';
		for(n = 0; n < sz; n += 4096)
			if(p[n] != 'B') {
				char c = '.';
				if(isprint(p[n]))
					c = p[n];
				trace_cerr << "Child: comparison failed at " << n << " : "
					"wanted 'B'(" << (int)'B' << "), got '" << c << "'("
					<< (int)c << ")" << endl;
				res = 1;
				break;
			}
		my_free(p);
		trace << "Child: calling mcheck()." << endl;
		mcheck(0);
		trace << "Child: Success.\n" << endl;
	}
	if(res)
		return res;
	if(argc < 3) {
		const char * sz;
		if(argc < 2)
			sz = "1";
		else
			sz = argv[1];
		trace << "Calling self with execl()." << endl;
		errno = 0;
		int status = execl(argv[0], argv[0], sz, "again", (char *)0);
		res = !WIFEXITED(status);
		trace_cerr << "execl() exited " << (res ? "abnormally" : "normally")
			<< " with status " << WEXITSTATUS(status) << "." << endl;
		if(WIFSIGNALED(status)) {
			trace_cerr << "execl() exited with signal "
				<< strsignal(WTERMSIG(status)) << "(" << WTERMSIG(status)
				<< ")." << endl;
		}
		if(status == -1) {
			trace_cerr << "system() exited with errno " << strerror(errno)
				<< "(" << errno << ")." << endl;
		}
		return res;
	}
	return 0;
}
