/* Copyright (C) 2011  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#define FUSE_USE_VERSION 26
#include <fuse.h>

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <signal.h>

#include <v3c/mutex.hpp>
#include <v3c/file-lock.h>
#include <v3c/trace_scope.hpp>
#include <v3c/string_utils.hpp>
#include <v3c/filesystem.hpp>
#include <v3c/smart_ptr.hpp>
#include <v3c/dump-mem.h>
#include <v3c/atomic.h>
#include <v3c/memmap.hpp>

#include <pthread.h>

#include <map>
#include <set>
#include <sstream>
#include <string>

#include "treedb-malloc.h"

#include <execinfo.h> // backtrace()

//#define traceY traceX
//#define TraceWrapY(x) TraceWrapX(x)
#define traceY trace
#define TraceWrapY(x) TraceWrap(x)

#define raw_dump(s) ::write(g_trace_fd, s.c_str(), s.length())

using namespace std;
using namespace v3c;
#if 0
struct LockScoper
{
	inline LockScoper(Mutex &m, const char * file, int line, bool bTrace)
	: m_m(m)
	, m_file(file)
	, m_line(line)
	, m_bTrace(bTrace)
	{
		if(bTrace) {
			trace_ostream(m_file, m_line) << m_name << '(' << pthread_self()
				<< ") {" << endl;
			trace_indent(true);
		}
		timespec tm = {1, 0}; // 1 second.
		while(!m_m.lock(tm)) {
			if(bTrace)
				trace << "lock scope blocked by " << lkpid << " ..." << endl;
		}
		lkpid = pthread_self();
	}
	inline ~LockScoper()
	{
		if(m_bTrace) {
			trace_indent(false);
			trace_ostream(m_file, m_line) << m_name << '(' << lkpid
				<< ") }" << endl;
		}
		lkpid = 0;
		m_m.unlock();
	}
	Mutex & m_m;
	static const char * m_name;
	const char * m_file;
	int m_line;
	bool m_bTrace;
	static intptr_t lkpid;
};
const char * LockScoper::m_name = "LockScope";
intptr_t LockScoper::lkpid = 0;

  #define LockScope(m) LockScoper _lsc(m, __FILE__,__LINE__, true)
  #define QuietLockScope(m) LockScoper _lsc(m, __FILE__,__LINE__, false)
#else
  #define LockScope(m)
  #define QuietLockScope(m)
#endif
static string trace_filename = fs::pwd() + string(1, DIR_SEP_CHAR)
	+ string("tmd-trace.txt");
static int g_trace_fd = -1;

void clean_up(int do_exit);

void close_trace_fd()
{
	close(g_trace_fd);
}

static size_t trace_file_length()
{
	struct stat st;
	st.st_size = 0;
	lstat(trace_filename.c_str(), &st);
	return st.st_size;
}

// Just for fun :)
static const char *hello_str = "Hello World!\n";

#define CONTROL_FD (size_t)-1
#define HELLO_FD (size_t)-2

// We can't use TRACE_FD as we store the programs PID in fi->fh so we can flush
// any remaining trace messages for that program when it closes the trace file.
//#define TRACE_FD (size_t)-2

static const char *control_path = "/control";
static const char *trace_path = "/trace";
static const char *hello_path = "/hello";


struct ClientContext;
struct Backing;
struct Overlay;
#define my_mkstemp(tag) \
  v3c::fs::mkstemp(tag, (const char *)0, "TREEDB_FUSE_TMPDIR")

typedef smart_ptr< ClientContext > ClientContextPtr;
typedef map< unsigned int, ClientContextPtr > ClientContextMap;
typedef map< unsigned int, unsigned int > ForkMap;

// We control access and update to the maps with this mutex.
v3c::Mutex g_mutex;

ClientContextMap g_contexts;
ForkMap g_forks;

static string short_name(const string & in)
{
	string::size_type pos = in.find_last_of('/');
	if(pos == string::npos)
		pos = 0;
	else
		++pos;
	return in.substr(pos);
}

struct ClientContext {
	enum Type { eBacking, eOverlay };
	ClientContext(Type t, ClientContext * parent_, unsigned int id_
		, int fd1_, const string & name1_, int fd2_, const string & name2_
		, int mode_)
	: m_type(t)
	, m_parent(parent_)
	, m_refcnt(1)
	, m_id(id_)
	, m_old_id(id_)
	, m_fd1(fd1_)
	, m_fd2(fd2_)
	, m_name1(name1_)
	, m_name2(name2_)
	, m_mode(mode_)
	, m_anonymous(false)
	, m_written_to(false)
	{
		trace << "ClientContext ctor " << (void *)this << '.' << endl;
	}
	virtual ~ClientContext()
	{
		if(m_fd1 == -1)
			return;
		close(m_fd1);
		unlink(m_name1.c_str());
		close(m_fd2);
		unlink(m_name2.c_str());
		trace << "ClientContext dtor " << (void *)this << '.' << endl;
	}
	const char * type_str() const
	{
		if(m_type == eBacking)
			return "backing";
		return "overlay";
	}
	string desc() const
	{
		string ret = string_inline_printf("%u(%u): type %s"
			, m_id, m_old_id, /*short_name(m_name1).c_str(),*/ type_str());
		return ret;
	}
	void AddRef();
	void Release();
	size_t read(char *buf, size_t size, off_t offset);
	size_t write(const char *buf, size_t size, off_t offset);
	ClientContext * read_context_for(size_t offset);
	ClientContext * next_read_context_for(size_t offset)
	{
		if(!m_parent) {
				trace << "ClientContext::next_read_context_for() : "
					"No next context!" << endl;
			return 0;
		}
		return m_parent->read_context_for(offset);
	}
	Type m_type;
	ClientContext * m_parent;
	unsigned int m_refcnt, m_id, m_old_id;
	mutable int m_fd1, m_fd2;
	string m_name1, m_name2;
	int m_mode;
	bool m_anonymous: 1;
	bool m_written_to: 1;
private:
	ClientContext(const ClientContext & o);
	ClientContext & operator =(const ClientContext & o);
};

ostream & operator <<(ostream & s, const ClientContextMap & m)
{
	ClientContextMap::const_iterator i = m.begin();
	ClientContextMap::const_iterator j = m.end();
	s << "count : " << m.size() << '\n';
	size_t n;
	for(n = 0; i != j; ++i, ++n)
		s << '[' << n << "] " << i->first << " -> " << i->second->desc() << '\n';
	s.flush();
	return s;
}

ClientContext * ClientContext::read_context_for(size_t offset)
{
	TraceWrapX("ClientContext::read_context_for(" << offset << ')');
	if((size_t)offset > OVERLAY_PAGES_SIZE) {
		trace << "ERROR: offset too large!" << endl;
		return 0;
	}
	ClientContext * This = this;
	offset /= TREEDB_PAGE_SIZE;
	unsigned int bit = offset & 0x03; // 0-7
	uint8_t mask = 1 << bit;
	offset /= 8;
	uint8_t bits = 0;
	while(This) {
		errno = 0;
		(void)pread(This->m_fd2, &bits, 1, offset);
		if(bits)
			trace << "read " << (uint32_t)bits << " from " << This->desc()
				<< " at offset " << offset << '.' << endl;
		if(bits & mask)
			return This;
		This = This->m_parent;
	}
	return 0; // Don't read rubbish from disk - memset it to zeros.
}

struct Backing : ClientContext {
/*	Backing(unsigned int id_)
	: ClientContext(ClientContext::eBacking, id_)
	{
	}*/
	Backing(unsigned int id_
		, int fd1_, const string & name1_, int fd2_, const string & name2_
		, int mode_)
	: ClientContext(ClientContext::eBacking, 0, id_
		, fd1_, name1_, fd2_, name2_, mode_)
	{
	}
	//ClientContext * read_context_for(size_t) { return this; }
private:
	Backing(const Backing & o);
	Backing & operator =(const Backing & o);
};

struct AnonContexts {
	AnonContexts()
	: m_entries()
	, m_key(0)
	{
	}
	ClientContext * anonymize(ClientContext * p)
	{
		TraceWrap("AnonContexts::anonymize(" << p->m_id << ')');

		ClientContextMap::iterator i = g_contexts.find(p->m_id);
		if(i == g_contexts.end()) {
			trace << p->desc() << " not found in g_contexts." << endl;
			return 0;
		}
		i->second.manage(0); // smart_ptr, don't delete p.
		g_contexts.erase(i);

		m_entries[m_key] = ClientContextPtr(p);
		trace << "New key is " << m_key << '.' << endl;
		p->m_id = m_key; // So it can erase itself.
		p->m_anonymous = true;
		p->m_refcnt = 0; // Will be AddRef'd twice.
		++m_key;
		return p;
	}

	ClientContextMap m_entries;
	unsigned int m_key;
} g_anonymous_contexts;

struct Overlay : ClientContext {
/*	Overlay(unsigned int id_)
	: ClientContext(ClientContext::eOverlay, id_)
	{}*/
	Overlay(ClientContext * parent, unsigned int id_
		, int fd1, const string & name1, int fd2, const string & name2)
	: ClientContext(ClientContext::eOverlay, parent, id_
		, fd1, name1, fd2, name2, parent->m_mode)
	{
		m_parent->AddRef();
	}

	//ClientContext * read_context_for(size_t offset);
private:
	Overlay(const Overlay & o);
	Overlay & operator =(const Overlay & o);
};

// Overlays map back onto a backing or another overlay.
// When a process g_forks again, it backs on to another overlay.
// Two g_overlays can't key by the same pid, so the original one becomes anonymous
// and can't be keyed by pid any more, so it's keyed by a monotonically
// increasing key.

ClientContext * context_for_pid(pid_t pid)
{
	TraceWrap("context_for_pid(" << pid << ')');
	ClientContextMap::iterator i = g_contexts.find(pid);
	if(i == g_contexts.end())
		return 0;
	return i->second.pointer();
}

ClientContext * anonymize(ClientContext * p)
{
	TraceWrap("anonymize(" << p->desc() << ')');
	p = g_anonymous_contexts.anonymize(p);
	return p;
}

void ClientContext::AddRef()
{
	TraceWrap("ClientContext::AddRef(" << desc() << ')');
	trace << m_refcnt << " -> " << (m_refcnt + 1) << endl;
	++m_refcnt;
	if(m_parent)
		m_parent->AddRef();
}

// TODO If this is an overlay and neither the parents overlay nor the childs
// overlay were written to, un-anonymize and delete the one remaining overlay.
void ClientContext::Release()
{
	TraceWrap("ClientContext::Release(" << desc() << ')');
	trace << m_refcnt << " -> " << (m_refcnt - 1) << endl;
	if(!m_refcnt) {
		trace << "Zero refcnt!" << endl;
		return; // ?
	}
	if(--m_refcnt)
		return; // Not zero yet.
	unsigned int id = m_id;
	trace << (m_anonymous ? "hidden" : "") << ' '
		<< ((m_type == ClientContext::eBacking) ? "backing" : "overlay")
		<< " context " << id << " going down." << endl;
	ClientContext * parent = m_parent;
	// One of these deletes "this".
	if(m_anonymous) {
		TraceWrap("hidden " << g_anonymous_contexts.m_entries.size());
		trace << g_anonymous_contexts.m_entries << flush;
		ClientContextMap::iterator i = g_anonymous_contexts.m_entries.find(id);
		if(i != g_anonymous_contexts.m_entries.end())
			g_anonymous_contexts.m_entries.erase(i);
		else{
			trace << "id " << id << " not found in g_contexts!" << endl;
		}
	}else{
		TraceWrap("visible " << g_contexts.size());
		trace << g_contexts << flush;
		ClientContextMap::iterator i = g_contexts.find(id);
		if(i != g_contexts.end())
			g_contexts.erase(i);
		else{
			trace << "id " << id << " not found in g_contexts!" << endl;
		}
	}
	if(parent) {
		TraceWrap("parent->Release()");
		parent->Release();
	}
}

size_t ClientContext::read(char *buf, size_t size, off_t ofoffset)
{
	size_t offset = (size_t)0 + ofoffset;
	TraceWrapY("ClientContext::read(" << short_name(m_name1) << ", " << offset
		<< ", " << size << ')');

	trace << "read" << endl;
	//raw_dump(dump_to_string(buf + offset, size, true));

	return pread(m_fd1, buf, size, offset);
}
static void ensure_size(const string & name, int fd, off_t sz)
{
	struct stat st;
	if(fstat(fd, & st) < 0)
		trace << "ensure_size : fstat \"" << name << "\" failed : "
			<< strerror(errno) << '.' << endl;
	if(st.st_size < sz)
		if(ftruncate(fd, sz) < 0)
			trace << "ensure_size : ftruncate \"" << name << "\" failed : "
				<< strerror(errno) << '.' << endl;
}

size_t ClientContext::write(const char *buf, size_t size, off_t offset)
{
	off_t up_size = V3C_ROUND_UP_T2(size_t, offset + size
		, TREEDB_PAGE_SIZE);
	TraceWrapY("ClientContext::write ");
	m_written_to = true;

	ensure_size(m_name1.c_str(), m_fd1, up_size);

	trace << "Data: " << short_name(m_name1) << ": fd " << m_fd1
		<< ", offset " << offset << ", size " << size
		<< ", offset+size " << (offset + size)
		<< " -> " << up_size << '.' << endl;

	v3c::MemMap mm1(m_fd1, PROT_READ | PROT_WRITE
		, MAP_SHARED, 0, up_size);
	if(!mm1.good()) {
		trace << "MemMap failed, errno is " << strerror(errno) << '.' << endl;
		return 0;
	}
	//raw_dump(dump_to_string(buf + offset, size, true));
	memcpy(mm1.data() + offset, buf, size);
	// Mark this page as the one to read from.
	offset /= TREEDB_PAGE_SIZE;
	up_size = V3C_ROUND_UP_T2(size_t, offset / 8 + 1, TREEDB_PAGE_SIZE);

	ensure_size(m_name2.c_str(), m_fd2, up_size);

	trace << "Bits: " << short_name(m_name2) << ": fd " << m_fd2 << ", size "
		<< (offset / 8 + 1) << " -> " << up_size << '.' << endl;

	v3c::MemMap mm2(m_fd2, PROT_READ | PROT_WRITE
		, MAP_SHARED, 0, up_size);
	if(!mm2.good()) {
		trace << "MemMap failed, errno is " << strerror(errno) << '.' << endl;
		return 0;
	}
	unsigned int bit = offset & 0x03; // 0-7
	offset /= 8;
	uint8_t bits = v3c_sync_fetch_and_or((uint8_t *)(mm2.data() + offset)
		, (uint8_t)(1 << bit));
	trace << "wrote " << (uint32_t)bits << " to " << desc() << "  at offset "
		<< offset << '.' << endl;
	return size;
}

static int tmd_getattr(const char *path, struct stat *stbuf)
{
	TraceWrap(__PRETTY_FUNCTION__);
	// KDE loves to call this.
	trace << "getattr(" << path << ", " << stbuf << ')' << endl;

	memset(stbuf, 0, sizeof(struct stat));
	if(strcmp(path, "/") == 0) {
		stbuf->st_mode = S_IFDIR | 0755;
		stbuf->st_nlink = 2;
		return 0;
	}
	if (strcmp(path, hello_path) == 0) {
		stbuf->st_mode = S_IFREG | 0444;
		stbuf->st_nlink = 1;
		stbuf->st_size = strlen(hello_str);
		trace << "hello" << endl;
		return 0;
	}
	if (strcmp(path, trace_path) == 0) {
		stbuf->st_mode = S_IFREG | 0600;
		stbuf->st_nlink = 1;
		stbuf->st_size = trace_file_length();
		return 0;
	}
	if(strcmp(path, control_path) == 0) {
		stbuf->st_mode = S_IFREG | 0200; // read-write
		stbuf->st_nlink = 1;
		stbuf->st_size = 0;
		return 0;
	}
	unsigned int ppid, pid;
	if(sscanf(path + 1, "%d-%d", &ppid, &pid) != 2) {
		traceX << "Not an overlay." << endl;
		ppid = 0;
		if(sscanf(path + 1, "%d", &pid) != 1) {
			traceX << "Not a backing." << endl;
			return -ENOENT;
		}
	}

	LockScope(g_mutex); // GLOBAL LOCK.

	ClientContextMap::iterator i = g_contexts.find(pid);
	if(i == g_contexts.end()) {
		trace << "pid " << pid << " not in g_contexts." << endl;
		if(ppid) {
			ForkMap::iterator i = g_forks.find(ppid);
			if(i == g_forks.end()) {
				traceX << "ppid " << ppid << " not in g_forks." << endl;
			}else{
				trace << "ppid " << ppid << " found in g_forks." << endl;
				stbuf->st_mode = S_IFREG | 0444;
				stbuf->st_nlink = 1;
				stbuf->st_size = OVERLAY_PAGES_SIZE;
				return 0;
			}
		}
		return -ENOENT;
	}
	trace << "pid " << pid << " found in g_contexts." << endl;
	stbuf->st_mode = S_IFREG | i->second->m_mode; // 0444;
	stbuf->st_nlink = 1;
	stbuf->st_size = OVERLAY_PAGES_SIZE;
	return 0;
}

static int tmd_readdir
( const char *path
, void *buf
, fuse_fill_dir_t filler
, off_t offset
, struct fuse_file_info *fi
) {
	TraceWrap("readdir " << path);

	(void) offset;
	(void) fi;

	if (strcmp(path, "/") != 0)
		return -ENOENT;

	filler(buf, ".", NULL, 0);
	filler(buf, "..", NULL, 0);

	LockScope(g_mutex); // GLOBAL LOCK.

	ClientContextMap::const_iterator i(g_contexts.begin());
	ClientContextMap::const_iterator j(g_contexts.end());
	for(; i != j; ++i) {
		stringstream ss;
		if(i->second->m_parent)
			ss << i->second->m_parent->m_id << '-' << i->first;
		else
			ss << i->first;
		filler(buf, ss.str().c_str(), NULL, 0);
	}
	filler(buf, control_path + 1, NULL, 0);
	filler(buf, hello_path + 1, NULL, 0);
	filler(buf, trace_path + 1, NULL, 0);
	return 0;
}

typedef map< pid_t, string > StringMap;
static StringMap trace_map;

static int tmd_create
( const char *path
, mode_t mode
, struct fuse_file_info *fi
);
static int tmd_release(const char *path, struct fuse_file_info *fi);

/** child overlay creation

1. insert tuple into g_forks
   if succeeded:
	2a. anonymize the parent context
	2b. overlay parent = anonymized parent context
   else: parent got here first after the fork
	2c. overlay parent = parents context parent (the anonymized context)
	2d. delete from g_forks
2. create overlay temp files
3. insert into g_overlays, keyed by pid
4. insert into g_contexts, keyed by pid
*/
static int tmd_open(const char *path, struct fuse_file_info *fi)
{
	TraceWrap("open " << path);

	if(!strcmp(path, control_path)) {
		fi->fh = CONTROL_FD;
		return 0;
	}
	if(!strcmp(path, trace_path)) {
		LockScope(g_mutex); // GLOBAL LOCK.
		pid_t pid = fuse_get_context()->pid;
		trace_map[pid] = "";
		fi->fh = pid; // not TRACE_FD; we need to know what to release.
		return 0;
	}
	if(!strcmp(path, hello_path)) {
		trace << "hello" << endl;
		if ((fi->flags & 3) != O_RDONLY) {
			trace << "(fi->flags & 3) != O_RDONLY" << endl;
			return -EACCES;
		}
		trace << "hello opened" << endl;
		fi->fh = HELLO_FD;
		return 0;
	}
	unsigned int ppid, pid, sres = sscanf(path + 1, "%d-%d", &ppid, &pid);
	if(sres != 2) {
		if(sres == 1) {
			trace << "The filesystem may be caching - will create." << endl;
			tmd_release(path, fi);
			return tmd_create(path, 0600, fi);
		}
		trace << "sscanf failed." << endl;
		return -ENOENT;
	}

	LockScope(g_mutex); // GLOBAL LOCK.

	trace << "Will create overlay for " << pid << '.' << endl;
	pair< ForkMap::iterator, bool > res
		= g_forks.insert( make_pair(ppid, pid) );

	ClientContext * fp;
	if(res.second) {
		// We just inserted the ppid
		trace << "Child: fork entry inserted." << endl;
		fp = anonymize(g_contexts[ppid].pointer());
	}else{
		// Our insert failed - it was already present. Delete it.
		trace << "Child: fork entry already inserted by " << ppid << '.'
			<< endl;
		fp = g_contexts[ppid]->m_parent;
		g_forks.erase(res.first);
	}
	trace << "Creating child overlay for " << pid << '.' << endl;
	// Create the overlay.
	string filename1 = "0-child-overlay1-XXXXXX";
	int fd1 = my_mkstemp(filename1);
	if(fd1 == -1)
		return -ENOENT;
	string filename2 = "0-child-overlay2-XXXXXX";
	int fd2 = my_mkstemp(filename2);
	if(fd2 == -1) {
		close(fd1);
		unlink(filename1.c_str());
		return -ENOENT;
	}
	Overlay * po = new Overlay(fp, pid, fd1, filename1, fd2, filename2);
	if(!po) {
		trace << "Out of memory creating overlay." << endl;
		//g_forks.erase(g_forks.find(ppid)); - parent may still succeed.
		close(fd1);
		unlink(filename1.c_str());
		close(fd2);
		unlink(filename2.c_str());
		return -ENOENT;
	}
	g_contexts[pid] = po;
	fi->fh = pid;
	trace << "Overlay inserted: " << po->desc() << '.' << endl;

	return 0;
}

/** parent backing creation

1. remove pid and temp files if the pid is present in g_contexts
2. create backing temp files
3. insert into g_backings, keyed by pid
4. insert into g_contexts, keyed by pid
*/
static int tmd_create
( const char *path
, mode_t mode
, struct fuse_file_info *fi
) {
	TraceWrap("tmd_create(" << path << ')');
	// Create a mapping for our client.
	unsigned int pid;
	if(sscanf(path + 1, "%d", &pid) != 1) {
		trace << "sscanf failed" << endl;
		return -1;
	}

	LockScope(g_mutex); // GLOBAL LOCK.

	ClientContext * fp = context_for_pid(pid);
	if(fp) {
		trace << "Discarding existing context." << endl;
		fp->Release();
	}
	string filename1 = "0-backing1-XXXXXX";
	int fd1 = my_mkstemp(filename1);
	if(fd1 == -1) {
		trace << "pid "<< pid << " : my_mkstemp backing 1 failed" << endl;
		return -1;
	}
	string filename2 = "0-backing2-XXXXXX";
	int fd2 = my_mkstemp(filename2);
	if(fd2 == -1) {
		trace << "pid "<< pid << " : my_mkstemp backing 2 failed" << endl;
		close(fd1);
		unlink(filename1.c_str());
		return -1;
	}
	fi->fh = pid;
	Backing * pb = new Backing(pid, fd1, filename1, fd2, filename2, mode);
	if(!pb) {
		trace << "Out of memory creating backing." << endl;
		close(fd1);
		unlink(filename1.c_str());
		close(fd2);
		unlink(filename2.c_str());
		return -ENOENT;
	}
	g_contexts[pid] = pb;
	return 0;
}

// NOTE: users cannot read from /control - it's write only
// TODO: make reading from /control dump some sort of status report?
static int tmd_read
( const char *path
, char *buf
, size_t size
, off_t offset
, struct fuse_file_info *fi
) {
	TraceWrapY("tmd_read(" << path << ", " << offset << ", " << size << ')');

	if(!strcmp(path, trace_path)) {
		int res = pread(g_trace_fd, buf, size, offset);
		return res;
	}

	if(!strcmp(path, hello_path)) {
		trace << "hello" << endl;
		off_t len = (off_t)strlen(hello_str);
		if (offset < len) {
			if ((off_t)(offset + size) > len)
				size = len - offset;
			memcpy(buf, hello_str + offset, size);
		} else
			size = 0;

		return size;
	}

	size_t original_size = size;
	uint64_t id = fi->fh;
	if(!id)
		return -ENOENT;

	ClientContext * fp0, *fp;
	{
		LockScope(g_mutex); // GLOBAL LOCK.
		ClientContextMap::iterator i = g_contexts.find(id);
		if(i == g_contexts.end())
			return -ENOENT;
		fp0 = i->second.pointer();
	}
	// Check for an odd block at the start.
	size_t odd = offset % TREEDB_PAGE_SIZE;
	size_t sz;
	if(odd) {
		// Read any odd-positioned chunk at the start.
		if((odd + size) < TREEDB_PAGE_SIZE)
			sz = size;
		else
			sz = TREEDB_PAGE_SIZE - odd;

		// Read any odd-sized chunk at the start.
		traceY << "Reading " << sz << " odd-size bytes of " << size
			<< " at offset " << offset << " from " << fp0->desc() << '.'
			<< endl;
		size_t res = fp0->read(buf, sz, offset);
		if(res != sz) {
			trace << "Error: reading " << sz << " bytes but read " << res
				<< " bytes." << endl;
			return res; // No other way to report errors.
		}

		traceY << "Done." << endl;

		// This will align to a page boundary.
		buf += sz;
		offset += sz;
		size -= sz;
	}
	sz = TREEDB_PAGE_SIZE;
	while(size >= TREEDB_PAGE_SIZE) {
		{
			LockScope(g_mutex); // GLOBAL LOCK.
			fp = fp0->read_context_for(offset);
		}
		if(fp) {
			traceY << "reading " << sz << " bytes of " << size << " from "
				<< fp->desc() << '.' << endl;
			size_t res = fp->read(buf, sz, offset);
			if(res != sz) {
				trace << "wanted " << sz << " bytes but read " << res
					<< " bytes." << endl;
				memset(buf + res, 0, sz - res);
			}
		}else{
			traceX << "Zeroing " << sz << " bytes of " << size
				<< " at offset " << offset << '.' << endl;
			// We could write something else here to stress test apps.
			memset(buf, 0, sz);
		}
		traceX << "Done." << endl;
		buf += sz;
		size -= sz;
		offset += sz;
	}
	if(size) {
		sz = size;
		{
			LockScope(g_mutex); // GLOBAL LOCK.
			fp = fp0->read_context_for(offset);
		}
		if(fp) {
			// Read any odd-sized chunk at the end.
			traceY << "Reading " << sz << " odd-size end bytes of " << size
				<< " from " << fp->desc() << '.' << endl;
			size_t res = fp->read(buf, sz, offset);
			if(res != sz) {
				traceY << "EOF reading " << sz << " bytes but read " << res
					<< " bytes." << endl;
				memset(buf + res, 0, sz - res);
			}
		}else{
			traceY << "Zeroing " << sz << " odd-size end bytes of " << size
				<< '.' << endl;
			// We could write something else here to stress test apps.
			memset(buf, 0, sz);
		}
		traceY << "Done." << endl;
	}
	return original_size;
}

void trace_remove_entry(pid_t pid)
{
	QuietLockScope(g_mutex); // GLOBAL LOCK.

	StringMap::iterator i = trace_map.find(pid);
	if(i == trace_map.end())
		return;
	string & text = i->second;
	if(text.length()) {
		text += '\n';
		string pid_str = string_inline_printf("[%zu] ", (size_t)pid);
		(void)write(g_trace_fd, pid_str.c_str(), pid_str.length());
		(void)write(g_trace_fd, text.c_str(), text.length());
	}
	trace_map.erase(i);
}
/**
We prepend the process id of the caller to complete trace lines.
Lines without a pid originate in this daemon.

We store partial writes in the trace_map, keyed by process id.
Complete lines are written when collected and at shutdown.
*/
int trace_write
( const char *buf
, size_t size
, pid_t pid
) {
	QuietLockScope(g_mutex); // GLOBAL LOCK.

/*	StringMap::iterator i = trace_map.find(pid);
	if(i == trace_map.end())
		return -ENOENT; // Not a pid that opened the file.
	string & text = i->second;*/
	string & text = trace_map[pid];
	text += string(buf, size);

	//(void)lseek(g_trace_fd, 0, SEEK_END);
	string::size_type pos = 0;
	string pid_str = string_inline_printf("[%zu] ", (size_t)pid);
	while(true) {
		pos = text.find_first_of('\n');
		if(pos == string::npos)
			break;
		++pos;
		(void)write(g_trace_fd, pid_str.c_str(), pid_str.length());
		(void)write(g_trace_fd, text.c_str(), pos);
		text = text.substr(pos);
	}
	return size;
}
/** parent calls "overlay my memory map"

1. insert tuple(ppid, 0) into g_forks as we don't know the childs pid yet
   if succeeded:
	2a. anonymize our context
	2b. overlay parent = anonymized context
   else: child got here first after the fork
	2c. overlay parent = our context
	2d. delete from g_forks
3. create overlay temp files
4. insert into g_overlays, keyed by ppid
5. insert into g_contexts, keyed by ppid
*/
int control_write
( const char *buf
, size_t size
) {
	TraceWrap("control_write");
	if(size < 1)
		return -ENOENT;
	if(buf[0] != 'O')
		return -ENOENT;
	int ppid, pid;
	if(sscanf(buf + 1, "%d", &ppid) != 1) {
		trace << "sscanf failed." << endl;
		return -ENOENT;
	}
	trace << "Control message: create overlay " << ppid << '.' << endl;

	LockScope(g_mutex); // GLOBAL LOCK.

	pair< ForkMap::iterator, bool > res
		= g_forks.insert( make_pair(ppid, 0) ); // We don't know the child pid.

	ClientContext * fp;
	if(res.second) { // We just inserted the ppid
		trace << "Fork entry inserted." << endl;
		fp = anonymize(g_contexts[ppid].pointer());
	}else{
		pid = res.first->second; // The child got there first.
		trace << "Fork entry already inserted by " << pid << '.' << endl;
		fp = g_contexts[pid]->m_parent; // Already anonymized.
		// Our insert failed - it was already present. Remove the entry.
		trace << "Removing fork entry." << endl;
		g_forks.erase(res.first);
	}
	trace << "Creating overlay." << endl;
	// Create the overlay.
	string filename1 = "0-parent-overlay1-XXXXXX";
	int fd1 = my_mkstemp(filename1);
	if(fd1 == -1) {
		trace << "my_mkstemp parent-overlay 1 failed." << endl;
		return -ENOENT;
	}
	string filename2 = "0-parent-overlay2-XXXXXX";
	int fd2 = my_mkstemp(filename2);
	if(fd2 == -1) {
		trace << "my_mkstemp parent-overlay 2 failed." << endl;
		close(fd1);
		unlink(filename1.c_str());
		return -ENOENT;
	}
	trace << "Created overlay " << short_name(filename1) << '.' << endl;

	Overlay * po = new Overlay(fp, ppid, fd1, filename1, fd2, filename2);
	if(!po) {
		trace << "Out of memory creating overlay." << endl;
		//g_forks.erase(g_forks.find(ppid)); - parent may still succeed.
		close(fd1);
		unlink(filename1.c_str());
		close(fd2);
		unlink(filename2.c_str());
		return -ENOENT;
	}
	g_contexts[ppid] = po;
	trace << "Overlay inserted: " << po->desc() << '.' << endl;
	// fi->fh = ppid; // done already.
	return size;
}

static int tmd_write
( const char *path
, const char *buf
, size_t size
, off_t offset
, struct fuse_file_info *fi
) {
	if(!size)
		return 0;
	(void)path;

	if(!strcmp(path, trace_path))
		return trace_write(buf, size, fi->fh);

	if(fi->fh == CONTROL_FD)
		return control_write(buf, size);

	TraceWrapY("tmd_write(" << path << ", " << offset << ", " << size << ')');

	size_t original_size = size;

	uint64_t id = fi->fh;
	if(!id || (id == HELLO_FD))
		return -ENOENT;

	ClientContext * fp;
	{
		LockScope(g_mutex); // GLOBAL LOCK.

		ClientContextMap::iterator i = g_contexts.find(id);
		if(i == g_contexts.end())
			return -ENOENT;
		fp = i->second.pointer();
	}
	// Check for an odd block at the start.
	size_t odd = offset % TREEDB_PAGE_SIZE;
	size_t sz;
	if(odd) {
		// Read any odd-positioned chunk at the start.
		if((odd + size) < TREEDB_PAGE_SIZE)
			sz = size;
		else
			sz = TREEDB_PAGE_SIZE - odd;

		traceY << "writing " << sz << " odd-size bytes of " << size
			<< " at offset " << offset << '.' << endl;
		size_t res;
		res = fp->write(buf, sz, offset);
		if(res != sz) {
			trace << "Error: writing " << sz << " bytes but wrote " << res
				<< " bytes." << endl;
			return res; // No other way to report errors.
		}
		traceY << "Done." << endl;

		// This is Copy On Write: copy the part of the page we didn't
		// write to from our parent.
		ClientContext * next;
		{
			LockScope(g_mutex); // GLOBAL LOCK.
			next = fp->next_read_context_for(offset);
		}
		char temp[odd];
		// Write to the start
		if(next) {
			traceY << "Reading head from parent." << endl;
			res = next->read(temp, odd, offset - odd);
		}else{
			traceY << "Zeroing out head." << endl;
			memset(temp, 0, odd);
		}
		fp->write(temp, odd, offset - odd);
		traceX << "Done." << endl;

		// This will align to a page boundary.
		buf += sz;
		offset += sz;
		size -= sz;
	}
	sz = TREEDB_PAGE_SIZE;
	while(size >= TREEDB_PAGE_SIZE) {
		traceY << "writing " << sz << " bytes of " << size << " at offset "
			<< offset << '.' << endl;
		size_t res;
		res = fp->write(buf, sz, offset);
		if(res != sz) {
			trace << "Error: writing " << sz << " bytes but wrote " << res
				<< " bytes." << endl;
			return res; // No other way to report errors.
		}
		traceY << "Done." << endl;
		offset += sz;
		buf += sz;
		size -= sz;
	}
	if(size) {
		sz = size;
		// Write any odd-sized chunk at the end.
		traceY << "writing " << sz << " odd-size bytes of " << size
			<< " at offset " << offset << '.' << endl;
		size_t res;
		res = fp->write(buf, sz, offset);
		if(res != sz) {
			trace << "Error: writing " << sz << " bytes but wrote " << res
				<< " bytes." << endl;
			return res; // No other way to report errors.
		}
		trace << "Done." << endl;
		// This is Copy On Write: copy the part of the page we didn't
		// write to from our parent.
		ClientContext * next;
		{
			LockScope(g_mutex); // GLOBAL LOCK.
			next = fp->next_read_context_for(offset);
		}
		char temp[TREEDB_PAGE_SIZE - size];
		// Write to the end
		if(next) {
			traceY << "Reading tail from parent." << endl;
			res = next->read(temp, offset + size, TREEDB_PAGE_SIZE - sz);
			if(res < TREEDB_PAGE_SIZE - sz)
				memset(temp, res, TREEDB_PAGE_SIZE - sz - res);
		}else{
			traceY << "Zeroing out tail." << endl;
			memset(temp, 0, TREEDB_PAGE_SIZE - sz);
		}
		fp->write(temp, TREEDB_PAGE_SIZE - size, offset + size);
		traceY << "Done." << endl;
	}
	return original_size;
}

static int tmd_release(const char *path, struct fuse_file_info *fi)
{
	TraceWrap("release(" << path << ')');
	(void)path;

	if(!strcmp(path, trace_path)) {
		if(fi->fh)
			trace_remove_entry(fi->fh);
		else
			trace << "Null fh!" << endl;
		return 0;
	}
	if(!strcmp(path, control_path))
		return 0;
	if (strcmp(path, hello_path) == 0)
		return 0;

	//trace << "will lock." << endl;
	LockScope(g_mutex); // GLOBAL LOCK.
	//trace << "locked." << endl;

	ClientContext * fp = context_for_pid(fi->fh);
	if(!fp) {
		trace << "ClientContext not found." << endl;
		return -ENOENT;
	}
	trace << "will Release() " << (void *)fp << '.' << endl;
	fp->Release();
	return 0;
}

/** Change the size of a file */
int tmd_truncate(const char *path, off_t size)
{
	TraceWrapX("truncate");
	(void)path;
	(void)size;
	return 1; // Nope.
}

/** Change the size of an open file */
int tmd_ftruncate(const char *path, off_t size, struct fuse_file_info * fi)
{
	(void)size;
	TraceWrapX("ftruncate(" << path << ", " << size << ')');
	(void)path;

	LockScope(g_mutex); // GLOBAL LOCK.

	// Only reduce the size. Let reads and writes increase it.
	ClientContextMap::iterator i = g_contexts.find(fi->fh);
	ClientContextMap::iterator j = g_contexts.end();

	if(i == j)
		return 1;
	return 0;
}
int tmd_fsync(const char *path, int isdatasync,
		     struct fuse_file_info *fi)
{
	/* Just a stub.	 This method is optional and can safely be left
	   unimplemented */

	(void) path;
	(void) isdatasync;
	(void) fi;
	return 0;
}

static struct fuse_operations tmd_oper;

const char * fuse_dir;
int created = 0;

void * tmd_init(struct fuse_conn_info *conn)
{
	(void) conn;
	trace << "init: pwd == " << v3c::fs::pwd() << endl;
	return 0;
}
void tmd_destroy(void *)
{
	TraceWrap("tmd_destroy");

	clean_up(0);
	// Delete it if we created it.
	if(created)
		rmdir(fuse_dir);
}

void clean_up_unlocked()
{
	TraceWrap("clean_up_unlocked(" << g_contexts.size() << " contexts)");
	trace << "cleaning up " << g_contexts.size() << " contexts and "
		<< g_anonymous_contexts.m_entries.size() << " anonymous contexts"
		<< endl;

	// Release the g_overlays.
	// The g_overlays reference anonymous or backing entries.
	// Once all the g_overlays are gone, the anonymous and backing entries will
	// be released too.
	g_contexts.clear();
	g_anonymous_contexts.m_entries.clear();
}
void clean_up(int do_exit)
{
	TraceWrap("clean_up(" << do_exit << ')');

	{
		LockScope(g_mutex); // GLOBAL LOCK.
		clean_up_unlocked();
	}
	if(do_exit)
		exit(0);
}

volatile sig_atomic_t fatal_error_in_progress = 0;

#if 0
/*
This isn't clean as fuse sets its own signal handlers.
TODO use the low-level interface and chain signal handlers to fuse.
For now, just rm 0-* in the TREEDB_FUSE_TMPDIR, TMPDIR, TMP or /tmp if you
have to kill this process (which you shouldn't have to :).
*/
#define STR_WRITE(name, txt) \
	static const char name[] = txt; \
	write(g_trace_fd, name, sizeof(name) - 1)
#define STATIC_STR_WRITE(txt) \
	write(g_trace_fd, txt, strlen(txt))
void signal_handler(int sig)
{
	/* Since this handler is established for more than one kind of signal,
		it might still get invoked recursively by delivery of some other kind
		of signal.  Use a static variable to keep track of that. */
	if (fatal_error_in_progress)
		raise (sig);
	fatal_error_in_progress = 1;

	/* This is pretty useless unless we can convert code locations to file: line
	void * buffer[50];
	int size = backtrace(buffer, 50);
	STATIC_STR_WRITE(a, "\nsignal handler called for ");
	STR_WRITE(strsignal(sig));
	STATIC_STR_WRITE(b, ". Backtrace:\n");
	backtrace_symbols_fd(buffer, size, g_trace_fd);
	*/

	if(sig != SIGSEGV) { // Too risky otherwise.
		if(g_trace_fd != -1)
			(void)lseek(g_trace_fd, 0, SEEK_END);
		clean_up_unlocked(); // Clean up what we can.
	}
	if(g_trace_fd != -1)
		close(g_trace_fd);

	/* Now reraise the signal.  We reactivate the signal's
		default handling, which is to terminate the process.
		We could just call `exit' or `abort',
		but reraising the signal sets the return status
		from the process correctly. */
	signal (sig, SIG_DFL);
	raise (sig);
}
#endif

#if V3C_DEBUG
class TmdWriterImpl
: public v3c::Tracer::Writer
{
public:
	inline TmdWriterImpl()
	: v3c::Tracer::Writer()
	{
	}
protected:
	virtual void write(const string & text)
	{
		string text2 = text + string(1, '\n');
		string::size_type len = text2.length();
		trace_write(text2.c_str(), len, 0);
	}
private:
	TmdWriterImpl(const TmdWriterImpl &);
	TmdWriterImpl & operator =(const TmdWriterImpl &);
};

typedef v3c::com::ReferencedObject< TmdWriterImpl, AutoMultiThreadModelTraits >
AutoTmdWriter;
typedef v3c::Tracer::GlobalWriter< AutoTmdWriter > TmdTraceWriter;
#endif

enum { KEY_HELP, KEY_VERSION };

static struct fuse_opt tmd_opts[] = {
	FUSE_OPT_KEY("--help", KEY_HELP),
	FUSE_OPT_KEY("--version", KEY_VERSION),
	FUSE_OPT_KEY("-h", KEY_HELP),
	FUSE_OPT_KEY("-V", KEY_VERSION),
/*	FUSE_OPT_KEY("stats", KEY_STATS),
	FUSE_OPT_KEY("cow", KEY_COW),
	FUSE_OPT_KEY("noinitgroups", KEY_NOINITGROUPS),
	FUSE_OPT_KEY("statfs_omit_ro", KEY_STATFS_OMIT_RO),
	FUSE_OPT_KEY("chroot=%s,", KEY_CHROOT),
	FUSE_OPT_KEY("max_files=%s", KEY_MAX_FILES),*/
	FUSE_OPT_END
};

struct uopt_t { int x; } uopt;

void uopt_init() {
	memset(&uopt, 0, sizeof(uopt_t)); // initialize options with zeros first
}

int tmd_opt_proc
( void *data
, const char *arg
, int key
, struct fuse_args *outargs
) {
	(void)data;
	(void)arg;
	(void)outargs;

	switch (key) {
		case FUSE_OPT_KEY_NONOPT:
/*			res = parse_branches(arg);
			if (res > 0) return 0;
			uopt.retval = 1;*/
			return 1;
// 		case KEY_STATS:
// 			uopt.stats_enabled = 1;
// 			return 0;
// 		case KEY_COW:
// 			uopt.cow_enabled = true;
// 			return 0;
// 		case KEY_STATFS_OMIT_RO:
// 			uopt.statfs_omit_ro = true;
// 			return 0;
// 		case KEY_NOINITGROUPS:
// 			// option only for compatibility with older versions
// 			return 0;
// 		case KEY_CHROOT:
// 			uopt.chroot = get_chroot(arg);
// 			return 0;
// 		case KEY_MAX_FILES:
// 			set_max_open_files(arg);
// 			return 0;
// 		case KEY_HELP:
// 			print_help(outargs->argv[0]);
// 			fuse_opt_add_arg(outargs, "-ho");
// 			uopt.doexit = 1;
// 			return 0;
// 		case KEY_VERSION:
// 			printf("unionfs-fuse version: "VERSION"\n");
// 			uopt.doexit = 1;
// 			return 1;
		default:
//			uopt.retval = 1;
			return 1;
	}
	return 0;
}

int main(int argc, char *argv[])
{
	struct fuse_args args = FUSE_ARGS_INIT(argc, argv);

	//CoutTraceWriter ctw;
	g_trace_fd = open(trace_filename.c_str(), O_RDWR | O_CREAT | O_TRUNC, 0600);
	if(g_trace_fd == -1) {
		cerr << "Error: couldn't open " << trace_filename << " for writing."
			<< endl;
		return 1;
	}
	atexit(close_trace_fd);

	trace_map[0] = ""; // Our slot.

	uopt_init();
	if (fuse_opt_parse(&args, NULL, tmd_opts, tmd_opt_proc) == -1) return 1;

	// Don't wait for the data to hit the disk, otherwise it crawls.
	if (fuse_opt_add_arg(&args, "-o")) {
		fprintf(stderr, "Severe failure, can't cache files(1), aborting!\n");
		return 2;
	}
	if (fuse_opt_add_arg(&args, "kernel_cache")) {
		fprintf(stderr, "Severe failure, can't cache files(2), aborting!\n");
		return 3;
	}
	// Disable multi-threading - it's still flaky.
	if (fuse_opt_add_arg(&args, "-s")) {
		fprintf(stderr, "Severe failure, can't disable multi-threading"
			", aborting!\n");
		return 2;
	}

#if V3C_DEBUG
	TmdTraceWriter tdw;
#endif

	TraceWrap("main");

	memset(&tmd_oper, 0, sizeof(fuse_operations));
	tmd_oper.flag_nullpath_ok = 0;
	tmd_oper.getattr = tmd_getattr;
	tmd_oper.readdir = tmd_readdir;
	tmd_oper.create = tmd_create;
	tmd_oper.open = tmd_open;
	tmd_oper.read = tmd_read;
	tmd_oper.write = tmd_write;
	tmd_oper.release = tmd_release;
	//tmd_oper.truncate = tmd_truncate;
	tmd_oper.ftruncate = tmd_ftruncate;
	tmd_oper.fsync = tmd_fsync;
	tmd_oper.init = tmd_init;
	tmd_oper.destroy = tmd_destroy;

// 	signal(SIGINT, signal_handler);
// 	signal(SIGHUP, signal_handler);
// 	signal(SIGTERM, signal_handler);
// 	signal(SIGSEGV, signal_handler);

	if(argc < 2) {
		cerr << argv[0] << " : fuse directory name not given." << endl;
		return 1;
	}
	int ret = fuse_main(args.argc, args.argv, &tmd_oper, NULL);
	trace << "treedb-malloc-daemon : result " << ret << '.' << endl;
	return ret;
}
