/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

/* THIS IS NOT A DISK BASED IMPLEMENTATION */

/**
A treedb_allocator provides an alternative to malloc, etc.

It does this by managing multiple 16 and 32 bit non-shared treedb's.
 - 16 bit treedb's store allcoations < 4K.\n
	These are stored in the 
 - 32 bit treedb's store allcoations >= 4K < 128K.
 - Anything larger gets its own mmap.

These "heap nodes" all have a treedb header, for consistency, along with tree
and list members so that they can be stored in a tree and in a list.

*/

#include "treedb.h"
#include <memory.h>
#include <stdlib.h>
#include <pthread.h>
#include "treedb-malloc-defines.h"

#include "treedb-malloc-impl.h"
#include <v3c/utf-16.h>

#ifdef __cplusplus
extern "C" {
#endif

char * treedb_malloc_get_base(malloc_heap_node_ptr_t p)
{
	switch(p->type) {
	case eMallocHeapNodeTree16:
		return p->a16.runtime.m_pBase;
	case eMallocHeapNodeTree32:
		return p->a32.runtime.m_pBase;
	case eMallocHeapNodeLarge:
		return p->large.runtime.m_pBase;
	default:
		return (char *)-1;
	}
}
#define ALLOCATOR_FREE_TREE(x) & (x)->runtime.m_pTreedb->m_allocator.m_free
#define ALLOCATOR_USED_TREE(x) & (x)->runtime.m_pTreedb->m_allocator.m_used
void * treedb_allocator_malloc_guarded
( treedb_allocator_runtime runtime
, size_t size
) {
	treedb_malloc_heap_node_ref_ptr_t t;
	if(size < runtime->small_allocation_threshold) {
		t.list_node = (L2(node_ptr_t))(runtime->a16l.m_head/*.node.value*/);
		A16(AvlAllocator_node_ptr_t) n;
		while(t.a16) {
			TreedbSearchResult res = AF16(search_ge)
				(& t.a16->runtime, ALLOCATOR_FREE_TREE(t.a16), size, & n);
			switch(res) {
			case eTreedbSearchEqual:
				AF16(unlink_node)(& t.a16->runtime, ALLOCATOR_FREE_TREE(t.a16), n);
				AU16(insert_node)(& t.a16->runtime, ALLOCATOR_USED_TREE(t.a16), n);
				return n->m_data;
			case eTreedbSearchGreater:
				if(n->m_uLength < (size + 32)) {
					AF16(unlink_node)(& t.a16->runtime, ALLOCATOR_FREE_TREE(t.a16), n);
					AU16(insert_node)(& t.a16->runtime, ALLOCATOR_USED_TREE(t.a16), n);
					return n->m_data;
				}
			default: /* eTreedbSearchEmpty, eTreedbSearchLess */
				break;
			}
			n = A16(alloc_node)(& t.a16->runtime, size);
			if(n) {
				AU16(insert_node)(& t.a16->runtime, ALLOCATOR_USED_TREE(t.a16), n);
				return n->m_data;
			}
			t.list_node = (L2(node_ptr_t))(t.list_node->m_next/*.node.value*/);
		}
		t.a16 = A16(new_heap)(runtime);
		if(t.a16) {
			L2(push_front_node)(& runtime->a16l, & runtime->a16l, t.list_node);
			HEAP(insert_node)(& runtime->heap, & runtime->heap, t.heap_node);
			n = A16(alloc_node)(& t.a16->runtime, size);
			if(n)
				return n->m_data;
			/* What? A16(alloc_node) should have worked! */
		}
	}
	/* Fall through if we fail. */
	if(size < runtime->medium_allocation_threshold) {
		t.list_node = (L2(node_ptr_t))(runtime->a32l.m_head/*.node.value*/);
		A32(AvlAllocator_node_ptr_t) n;
		while(t.a32) {
			TreedbSearchResult res = AF32(search_ge)
				(& t.a32->runtime, ALLOCATOR_FREE_TREE(t.a32), size, & n);
			switch(res) {
			case eTreedbSearchEqual:
				AF32(unlink_node)(& t.a32->runtime
					, ALLOCATOR_FREE_TREE(t.a32), n);
				AU32(insert_node)(& t.a32->runtime
					, ALLOCATOR_USED_TREE(t.a32), n);
				return n->m_data;
			case eTreedbSearchGreater:
				if(n->m_uLength < (size + 32 * 1024)) {
					AF32(unlink_node)(& t.a32->runtime
						, ALLOCATOR_FREE_TREE(t.a32), n);
					AU32(insert_node)(& t.a32->runtime
						, ALLOCATOR_USED_TREE(t.a32), n);
					return n->m_data;
				}
			default: /* eTreedbSearchEmpty, eTreedbSearchLess */
				break;
			}
			n = A32(alloc_node)(& t.a32->runtime, size);
			if(n) {
				AU32(insert_node)(& t.a32->runtime
					, ALLOCATOR_USED_TREE(t.a32), n);
				return n->m_data;
			}
			t.list_node = (L2(node_ptr_t))(t.list_node->m_next/*.node.value*/);
		}
		t.a32 = A32(new_heap)(runtime);
		if(t.a32) {
			L2(push_front_node)(& runtime->a32l, & runtime->a32l, t.list_node);
			HEAP(insert_node)(& runtime->heap, & runtime->heap, t.heap_node);
			n = A32(alloc_node)(& t.a32->runtime, size);
			if(n)
				return n->m_data;
			/* What? A32(alloc_node) should have worked! */
		}
	}
	t.large = LG(new_heap)(runtime, size);
	if(!t.large)
		return 0;
	HEAP(insert_node)(& runtime->heap, & runtime->heap, t.heap_node);
	return t.large->runtime.m_pData->data;
}
void * treedb_allocator_malloc(treedb_allocator_runtime runtime, size_t size)
{
	/* LOCK */
	int res = pthread_mutex_lock(& runtime->m_mx);
	if(res)
		return 0;
	void * p = treedb_allocator_malloc_guarded(runtime, size);
	/* UNLOCK */
	pthread_mutex_unlock(& runtime->m_mx);
	return p;
}
void * treedb_allocator_calloc(treedb_allocator_runtime runtime, size_t count
, size_t size)
{
	/* LOCK */
	int res = pthread_mutex_lock(& runtime->m_mx);
	if(res)
		return 0;
	size_t num = count * size;
	char * p = treedb_allocator_malloc_guarded(runtime, num);
	memset(p, '\0', num);
	/* UNLOCK */
	pthread_mutex_unlock(& runtime->m_mx);
	return p;
}
void * treedb_allocator_realloc
( treedb_allocator_runtime runtime
, void * p
, size_t new_size
) {
	if(!p)
		return treedb_allocator_malloc(runtime, new_size);
	/* LOCK */
	int res = pthread_mutex_lock(& runtime->m_mx);
	if(res)
		return 0;
	/* Find out which heap it's in. */
	treedb_malloc_heap_node_ref_ptr_t t;
	TreedbSearchResult sres
		= HEAP(search_le)(& runtime->heap, & runtime->heap, p, & t.heap_node);
	if(sres == eTreedbSearchGreater) {
		/* UNLOCK */
		pthread_mutex_unlock(& runtime->m_mx);
		return 0; /* It wasn't allocated from one of runtime's heaps. */
	}
	void * q;
	switch(* t.type) {
	case eMallocHeapNodeTree16: {
		if((((char *)p) - t.a16->runtime.m_pBase)
				>= (ssize_t)runtime->small_allocation_heap_size) {
			/* UNLOCK */
			pthread_mutex_unlock(& runtime->m_mx);
			return 0; /* It wasn't allocated from t. */
		}
		if((((char *)p) - t.a16->runtime.m_pBase)
				< (ssize_t)sizeof(A16(allocator_t))) {
			/* UNLOCK */
			pthread_mutex_unlock(& runtime->m_mx);
			return 0; /* It wasn't allocated from t. */
		}

		// Transfer the lock from the treedb allocator to the specific heap.
		res = pthread_mutex_lock(& t.a16->runtime.m_mx);
		if(res) {
			/* UNLOCK */
			pthread_mutex_unlock(& runtime->m_mx);
			return 0;
		}
		pthread_mutex_unlock(& runtime->m_mx);
		// From here, only the heap we found is locked.

		q = A16(realloc)(& t.a16->runtime, p, new_size);
		if(q) {
			/* UNLOCK */
			pthread_mutex_unlock(& t.a16->runtime.m_mx);
			return q;
		}
		/* Out of addresses or not enough space. */

		/* Back out of the lock. */
		pthread_mutex_unlock(& t.a16->runtime.m_mx);
		/* Re-enter the outer lock. */
		res = pthread_mutex_lock(& runtime->m_mx);
		if(res)
			return 0;

		A16(AvlAllocator_node_ptr_t) n1 = (A16(AvlAllocator_node_ptr_t))
			((char *)p
				- V3C_ROUND_UP_T2(uint16_t, sizeof(A16(AvlAllocator_node_ptr_t))
					, sizeof(uint16_t)));
		q = treedb_allocator_malloc_guarded(runtime, new_size);
		if(!q) {
			/* UNLOCK */
			pthread_mutex_unlock(& runtime->m_mx);
			return 0;
		}
		size_t len = n1->m_uLength;
		if(len > new_size)
			len = new_size;
		memcpy(q, p, len);
		A16(free)(& t.a16->runtime, p);
		/* UNLOCK */
		pthread_mutex_unlock(& runtime->m_mx);
		return q; }
	case eMallocHeapNodeTree32: {
		if((((char *)p) - t.a32->runtime.m_pBase)
				>= (ssize_t)runtime->medium_allocation_heap_size) {
			/* UNLOCK */
			pthread_mutex_unlock(& runtime->m_mx);
			return 0; /* It wasn't allocated from t. */
		}
		if((((char *)p) - t.a16->runtime.m_pBase)
				< (ssize_t)sizeof(A32(allocator_t))) {
			/* UNLOCK */
			pthread_mutex_unlock(& runtime->m_mx);
			return 0; /* It wasn't allocated from t. */
		}

		// Transfer the lock from the treedb allocator to the specific heap.
		res = pthread_mutex_lock(& t.a16->runtime.m_mx);
		if(res) {
			pthread_mutex_unlock(& runtime->m_mx);
			return 0;
		}
		pthread_mutex_unlock(& runtime->m_mx);
		// From here, only the heap we found is locked.

		q = A32(realloc)(& t.a32->runtime, p, new_size);
		if(q) {
			/* UNLOCK */
			pthread_mutex_unlock(& runtime->m_mx);
			return q;
		}
		/* Out of addresses or not enough space. */

		/* Back out of the lock. */
		pthread_mutex_unlock(& t.a16->runtime.m_mx);
		/* Re-enter the outer lock. */
		res = pthread_mutex_lock(& runtime->m_mx);
		if(res)
			return 0;

		A32(AvlAllocator_node_ptr_t) n1 = (A32(AvlAllocator_node_ptr_t))
			((char *)p
				- V3C_ROUND_UP_T2(uint32_t, sizeof(A32(AvlAllocator_node_ptr_t))
					, sizeof(uint32_t)));
		q = treedb_allocator_malloc_guarded(runtime, new_size);
		if(!q) {
			/* UNLOCK */
			pthread_mutex_unlock(& runtime->m_mx);
			return 0;
		}
		size_t len = n1->m_uLength;
		if(len > new_size)
			len = new_size;
		memcpy(q, p, len);
		A32(free)(& t.a32->runtime, p);
		/* UNLOCK */
		pthread_mutex_unlock(& runtime->m_mx);
		return q; }
	case eMallocHeapNodeLarge: {
		size_t len = t.large->runtime.m_pData->length;
		if((((char *)p) - t.large->runtime.m_pData->data)
				>= (ssize_t)(t.large->runtime.m_pData->length)) {
			/* UNLOCK */
			pthread_mutex_unlock(& runtime->m_mx);
			return 0; /* It wasn't t. */
		}
		if((((char *)p) - t.a16->runtime.m_pBase)
				< (ssize_t)sizeof(uint64_t)) {
			/* UNLOCK */
			pthread_mutex_unlock(& runtime->m_mx);
			return 0; /* It wasn't t. */
		}

		void * q;
		HEAP(unlink_node)(& runtime->heap, & runtime->heap, t.heap_node);
		if(new_size >= runtime->medium_allocation_threshold) {
			q = LG(realloc)(runtime, t.large, new_size);
			HEAP(insert_node)(& runtime->heap, & runtime->heap, t.heap_node);
			/* UNLOCK */
			pthread_mutex_unlock(& runtime->m_mx);
			return q; /* Return whether we succeeded or not. */
		}
		/* We're going to transfer this to a smaller heap. */
		q = treedb_allocator_malloc_guarded(runtime, new_size);
		if(!q) {
			LG(free_heap)(runtime, t.large);
			/* UNLOCK */
			pthread_mutex_unlock(& runtime->m_mx);
			return q; /* Return whether we succeeded or not. */
		}
		if(len > new_size)
			len = new_size;
		memcpy(q, p, len);
		LG(free_heap)(runtime, t.large);
		/* UNLOCK */
		pthread_mutex_unlock(& runtime->m_mx);
		return q; }
	default:
		/* UNLOCK */
		pthread_mutex_unlock(& runtime->m_mx);
		return 0;
	}
}
char * treedb_allocator_strdup(treedb_allocator_runtime runtime, const char * p)
{
	if(!p)
		return 0;
	size_t len = strlen(p);
	char * res = treedb_allocator_malloc(runtime, len);
	if(!res)
		return 0;
	strcpy(res, p);
	return res;
}
uint16_t * treedb_allocator_wcsdup
( treedb_allocator_runtime runtime
, const uint16_t * p
) {
	size_t len = utf16_strlen(p);
	uint16_t * res = (uint16_t *)
		treedb_allocator_malloc(runtime, len * sizeof(uint16_t));
	if(!res)
		return 0;
	utf16_strcpy(res, p);
	return res;
}
TreedbFreeResult treedb_allocator_free
( treedb_allocator_runtime runtime
, void * p
) {
	/* LOCK */
	int res = pthread_mutex_lock(& runtime->m_mx);
	if(res)
		return eTreedbFreeBadHeapLock;
	/* Find out which heap it's in. */
	treedb_malloc_heap_node_ref_ptr_t t;
	TreedbSearchResult sres = HEAP(search_le)(& runtime->heap, & runtime->heap, p
		, & t.heap_node);
	if(sres == eTreedbSearchGreater) {
		/* UNLOCK */
		pthread_mutex_unlock(& runtime->m_mx);
		/* It wasn't allocated from one of runtime's heaps. */
		return eTreedbFreeUnknownPointer;
	}
	switch(* t.type) {
	case eMallocHeapNodeTree16: {
		if((((char *)p) - t.a16->runtime.m_pBase)
				>= (ssize_t)runtime->small_allocation_heap_size) {
			pthread_mutex_unlock(& runtime->m_mx);
			/* UNLOCK */
			return eTreedbFreeUnknownPointer; /* It wasn't allocated from t. */
		}
		if((((char *)p) - t.a16->runtime.m_pBase)
				< (ssize_t)sizeof(A16(allocator_t))) {
			pthread_mutex_unlock(& runtime->m_mx);
			/* UNLOCK */
			return eTreedbFreeUnknownPointer; /* It wasn't allocated from t. */
		}
		TreedbFreeResult ret = A16(free)(& t.a16->runtime, p);
		if(runtime->free_emptied_heaps
			&& (t.a16->runtime.m_pTreedb->m_allocator.m_used.m_head/*.node.value*/ == 0)) {
			L2(unlink_node)(& runtime->a16l, & runtime->a16l, t.list_node);
			HEAP(unlink_node)(& runtime->heap, & runtime->heap, t.heap_node);
			A16(free_heap)(runtime, t.a16);
		}
		pthread_mutex_unlock(& runtime->m_mx);
		/* UNLOCK */
		return ret; }
	case eMallocHeapNodeTree32: {
		if((((char *)p) - t.a32->runtime.m_pBase)
				>= (ssize_t)runtime->medium_allocation_heap_size) {
			/* UNLOCK */
			pthread_mutex_unlock(& runtime->m_mx);
			/* It wasn't allocated from t. */
			return eTreedbFreeUnknownPointer;
		}
		if((((char *)p) - t.a16->runtime.m_pBase)
				< (ssize_t)sizeof(A32(allocator_t))) {
			/* UNLOCK */
			pthread_mutex_unlock(& runtime->m_mx);
			/* It wasn't allocated from t. */
			return eTreedbFreeUnknownPointer;
		}
		TreedbFreeResult ret = A32(free)(& t.a32->runtime, p);
		if(runtime->free_emptied_heaps
			&& (t.a32->runtime.m_pTreedb->m_allocator.m_used.m_head/*.node.value*/ == 0)) {
			L2(unlink_node)(& runtime->a32l, & runtime->a32l, t.list_node);
			HEAP(unlink_node)(& runtime->heap, & runtime->heap, t.heap_node);
			A32(free_heap)(runtime, t.a32);
		}
		/* UNLOCK */
		pthread_mutex_unlock(& runtime->m_mx);
		return ret; }
	case eMallocHeapNodeLarge: {
		if((((char *)p) - t.large->runtime.m_pData->data)
				>= (ssize_t)(t.large->runtime.m_pData->length)) {
			/* UNLOCK */
			pthread_mutex_unlock(& runtime->m_mx);
			/* It wasn't t. */
			return eTreedbFreeUnknownPointer;
		}
		if((((char *)p) - t.a16->runtime.m_pBase)
				< (ssize_t)sizeof(uint64_t)) {
			/* UNLOCK */
			pthread_mutex_unlock(& runtime->m_mx);
			/* It wasn't t. */
			return eTreedbFreeUnknownPointer;
		}
		HEAP(unlink_node)(& runtime->heap, & runtime->heap, t.heap_node);
		TreedbFreeResult ret = LG(free_heap)(runtime, t.large);
		/* UNLOCK */
		pthread_mutex_unlock(& runtime->m_mx);
		return ret; }
	default:
		pthread_mutex_unlock(& runtime->m_mx);
		return eTreedbFreeBadHeap;
	}
}

#ifdef __cplusplus
} // extern "C" {
#endif
