/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

/*! @page avl avl trees
An avl-tree implements the tree concept.

It consists of a root storing the optional count, optional depth and a packed
pointers to the tree head, tail and top.

A packed pointer (P1) is typically used to refer to the tree. If it's part of
the treedb "payload" then it's offset is at a fixed location - a constant
value.

This avl implementation was originally based on the work of
Michael H. Buselli <cosine@cosine.org>, Wessel Dankers <wsl@nl.linux.org>

and an alternate implementation written by Ben Pfaff <blp@gnu.org>.

It's sufficiently different that questions regarding the implementation should
be addressed to your's truely, Philip Ashmore <contact@philipashmore.com>.

avl.h and avl-impl.h are @ref c-templates.

@section setup What to do before including avl.h / avl-impl.h
Make sure your definitions before including the header are consistent with those
before including implementation.

The raw AVL_TMPL_omit* list was generated by
\code
cat v3c/avl-impl.h | \
grep AVL_TMPL_omit | \
sed -r -e 'sX^ *\t*\|* ?#?(if)?n?def(ine)?d? +XXg' \
	-e 's|#if \(defined ||g' \
	-e 's| \\$||g' \
| sort \
| uniq
\endcode

Defining the AVL_TMPL_omit*_DEFINED macro causes the AVL_TMPL_omit* macro to
be defined in @ref avl.h .

1. header - omit type declarations
 -# AVL_TMPL_omit_tree_t
 -# AVL_TMPL_omit_tree_t_typedefs
 -# AVL_TMPL_omit_node_t
 -# AVL_TMPL_omit_node_t_typedefs

2. header+implementation - omit function declarations+definitions
 -# AVL_TMPL_omit_clear_DECLARE
 -# AVL_TMPL_omit_clone_DECLARE
 -# AVL_TMPL_omit_delete_node_DECLARE
 -# AVL_TMPL_omit_distance_DECLARE
 -# AVL_TMPL_omit_free_DECLARE
 -# AVL_TMPL_omit_index_DECLARE
 -# AVL_TMPL_omit_init_DECLARE
 -# AVL_TMPL_omit_insert_DECLARE
 -# AVL_TMPL_omit_insert_node_DECLARE
 -# AVL_TMPL_omit_insert_node_after_DECLARE
 -# AVL_TMPL_omit_insert_node_before_DECLARE
 -# AVL_TMPL_omit_insert_node_top_DECLARE
 -# AVL_TMPL_omit_key_distance_DECLARE
 -# AVL_TMPL_omit_node_at_index_DECLARE
 -# AVL_TMPL_omit_node_distance_DECLARE
 -# AVL_TMPL_omit_node_index_DECLARE
 -# AVL_TMPL_omit_pack_node_DECLARE
 -# AVL_TMPL_omit_pack_tree_DECLARE
 -# AVL_TMPL_omit_remove_DECLARE
 -# AVL_TMPL_omit_remove_between_DECLARE
 -# AVL_TMPL_omit_remove_many_DECLARE
 -# AVL_TMPL_omit_search_DECLARE
 -# AVL_TMPL_omit_search_closest_DECLARE
 -# AVL_TMPL_omit_search_ge_DECLARE
 -# AVL_TMPL_omit_search_gt_DECLARE
 -# AVL_TMPL_omit_search_le_DECLARE
 -# AVL_TMPL_omit_search_lt_DECLARE
 -# AVL_TMPL_omit_unlink_node_DECLARE
 -# AVL_TMPL_omit_unpack_node_DECLARE
 -# AVL_TMPL_omit_unpack_tree_DECLARE

3. implementation - omit function definitions
 -# AVL_TMPL_omit_clear
 -# AVL_TMPL_omit_clone
 -# AVL_TMPL_omit_delete_node
 -# AVL_TMPL_omit_distance
 -# AVL_TMPL_omit_free
 -# AVL_TMPL_omit_index
 -# AVL_TMPL_omit_init
 -# AVL_TMPL_omit_insert
 -# AVL_TMPL_omit_insert_node
 -# AVL_TMPL_omit_insert_node_after
 -# AVL_TMPL_omit_insert_node_before
 -# AVL_TMPL_omit_insert_node_top
 -# AVL_TMPL_omit_key_distance
 -# AVL_TMPL_omit_node_at_index
 -# AVL_TMPL_omit_node_distance
 -# AVL_TMPL_omit_node_index
 -# AVL_TMPL_omit_pack_node
 -# AVL_TMPL_omit_pack_tree
 -# AVL_TMPL_omit_remove
 -# AVL_TMPL_omit_remove_between
 -# AVL_TMPL_omit_remove_many
 -# AVL_TMPL_omit_search
 -# AVL_TMPL_omit_search_closest
 -# AVL_TMPL_omit_search_ge
 -# AVL_TMPL_omit_search_gt
 -# AVL_TMPL_omit_search_le
 -# AVL_TMPL_omit_search_lt
 -# AVL_TMPL_omit_unlink_node
 -# AVL_TMPL_omit_unpack_node
 -# AVL_TMPL_omit_unpack_tree

4. must-defines
 -# AVL_TMPL_NS(x)
	Every type and function is wrapped using this macro to allow for multiple
	avl tree "types", i.e. string trees and int trees.

5. types : type [ default ]
 -# AVL_TMPL_runtime_ptr_t [ AVL_TMPL_NS(runtime_ptr_t) ]\n
	This is the type of the collection and is always named "runtime".\n
	This way, the internal macros don't need to be passed this as a parameter.
 -# AVL_TMPL_size_t [ uintptr_t ]\n
	The type used to exchange scalar information with the client.
 -# AVL_TMPL_ssize_t [ intptr_t ]\n
	The type used to exchange signed scalar information with the client.
 -# AVL_TMPL_count_t [ AVL_TMPL_size_t ]\n
	The type used in the implementation for scalars and packed pointers.
 -# AVL_TMPL_tree_ptr_t [ AVL_TMPL_NS(tree_ptr_t) ]
 -# AVL_TMPL_node_ptr_t [ AVL_TMPL_NS(node_ptr_t) ]
 -# AVL_TMPL_packed_tree_ptr_t [ AVL_TMPL_count_t ]
 -# AVL_TMPL_packed_node_ptr_t [ AVL_TMPL_count_t ]
 -# AVL_TMPL_packed_ptr_t [ AVL_TMPL_count_t ]
 -# AVL_TMPL_key_t [ AVL_TMPL_data_t ]\n
	Used for comparison.\n
	The type of the item member used for searching/sorting, or the whole item,
	for collections like sets.
 -# AVL_TMPL_data_t [ AVL_TMPL_NS(data_t) ]\n
	The payload of the node. It may just be a key, but could also be a struct.
	Any "pointers" it contains to allocated data must be packed.
 -# AVL_TMPL_input_t [ AVL_TMPL_data_t ]\n
	Used for insertion. Contains enough data to decide
	 - if the item to be inserted is unique
	 - the required size of the containing node
	It could contain all the information required to polulate a data_t, but for
	complex data_t's it's quicker and easier to just supply enough data for the
	allocation + insertion to complete and then polulate the data_t directly.

6. optional : name [ default ]
 -# AVL_TMPL_prefix [ extern ]\n
	Prefixes function prototypes/definitions.\n
	A sometimes useful alternative is [ static ]
 -# AVL_TMPL_has_count [ 1 ] and/or AVL_TMPL_has_depth [ 1 ]\n
	Specifies if a node has a count and/or depth members.
	Either can be used for balancing, but having both is fastest.
	Omitting both implies both

7. header - constants\n
 -# AVL_TMPL_init [ { 0, 0, 0 } ]

8. avl tree structure members : var [ default ]
 -# AVL_TMPL_head [ m_head ]
 -# AVL_TMPL_tail [ m_tail ]
 -# AVL_TMPL_top [ m_top ]

9. avl tree node structure members : var [ default ]\n
	Occasionally, one may need to store the same data in more than one
	container. Renaming the node members is sometimes required.
 -# AVL_TMPL_next [ m_next ]
 -# AVL_TMPL_prev [ m_prev ]
 -# AVL_TMPL_parent [ m_parent ]
 -# AVL_TMPL_left [ m_left ]
 -# AVL_TMPL_right [ m_right ]
 -# AVL_TMPL_head [ m_head ]
 -# AVL_TMPL_tail [ m_tail ]
 -# AVL_TMPL_top [ m_top ]
 -# AVL_TMPL_count [ m_count ]
 -# AVL_TMPL_depth [ m_depth ]
 -# AVL_TMPL_item [ m_item ] This is of type data_t.
 -# AVL_TMPL_item_SUFFIX [ ]
	While not a member, strictly speaking, this is sometimes usefully defined as
	"[]" to denote an unbounded array, causing m_item to be of type data_t[].

10. implementation : must-defines\n
 -# AVL_TMPL_free(n) to free a tree root\n
	Required if AVL_TMPL_omit_free(_DECLARE) aren't defined.
 -# AVL_TMPL_alloc_node(i) to allocate a node
 -# AVL_TMPL_free_node(n) to free a node
 -# AVL_TMPL_unpack_pointer(u) to unpack a pointer\n
	Unpacking/packing a pointer converts to/from the internal representation of
	the pointer. Usually this internal representation is an offset from the
	start of the "heap".
 -# AVL_TMPL_pack_pointer(p) to pack a pointer
 -# AVL_TMPL_cmp(a,b) to compare a with b, like strcmp()

11. implementation : optional : var [ default ]\n
 -# AVL_TMPL_free_item_data(pn) [ ]\n
	to free any allocated data associated with the node.
 -# AVL_TMPL_unpack_key(pn) [ (pn)->AVL_TMPL_item ]\n
	extract the "key" from the supplied data_t
 -# AVL_TMPL_free_packed_tree(x) [ AVL_TMPL_free(AVL_TMPL_unpack_tree(x)) ]
 -# AVL_TMPL_free_tree_pointer(x) [ AVL_TMPL_free(x) ]
 -# AVL_TMPL_free_packed_node(x) [ AVL_TMPL_free_node(AVL_TMPL_unpack_node(x)) ]
 -# AVL_TMPL_free_node_pointer(x) [ AVL_TMPL_free_node(x) ]
 -# AVL_TMPL_allow_duplicates [ 0 ]\n
	Decides how to handle duplicates.
 -# AVL_TMPL_clone_node(pn) [ AVL_TMPL_alloc_node(pn->AVL_TMPL_item) ]\n
	This could be a bad choice if the input and stored representations differ.
 -# AVL_TMPL_exceeds_count_limit(u)\n
	Tests if the value exceeds the range of the count type.

12. implementation : insertion / extraction : var [ default ]\n
 -# AVL_TMPL_unpack_tree(x) [ AVL_TMPL_NS(unpack_tree)(runtime, x) ]
 -# AVL_TMPL_unpack_node(x) [ AVL_TMPL_NS(unpack_node)(runtime, x) ]
 -# AVL_TMPL_pack_tree(x) [ AVL_TMPL_NS(pack_tree)(runtime, x) ]
 -# AVL_TMPL_pack_node(x) [ AVL_TMPL_NS(pack_node)(runtime, x) ]

@section helper avl-helper.h et al
Some (more) boilerplate code.
*/
