/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

/*! @page treedb-handle-array handle arrays

A treedb-handle-array (handle-array) implements the array concept.

It consists of a root storing the location of the array state.

The array state consists of a header storing the count, reserve amount,
allocation increment followed by the array data.

A packed pointer (P1) is maintained so that it always refers to the location of
the array state, which can move if reallocated.

As this is at a fixed location(offset) P1's offset (P2) can be passed around as
its value doesn't change.

As the treedb itself may be resized and move as a result of an array resize, any
functions that could resize the array must be passed P2.

@section setup What to do before including header / implementation
Make sure your definitions before including the header are consistent with those
before including implementation.

1. header - omit type declarations
 -# ARRAY_omit_array_t
 -# ARRAY_omit_array_t_typedefs

2. header+implementation - omit function declarations+definitions
 -# ARRAY_omit_new_DECLARE
 -# ARRAY_omit_free_DECLARE

3. implementation - omit function definitions
 -# ARRAY_omit_new
 -# ARRAY_omit_free

4. header - must-defines\n
 -# ARRAY_NS(x)
	Every type and function is wrapped using this macro to allow for multiple
	array "types", i.e. float arrays and int arrays.

5. header/implementation - types : type [ default ]\n
 -# ARRAY_runtime_ptr_t [ ARRAY_NS(runtime_ptr_t) ]\n
	This is the type of the array and is always named "runtime".\n
	This way, the internal macros don't need to be passed this as a parameter.
 -# ARRAY_size_t [ uintptr_t ]\n
	The type used to exchange scalar information with the client.
 -# ARRAY_count_t [ ARRAY_size_t ]\n
	The type used in the implementation for scalars and packed pointers.
 -# ARRAY_ptr_t [ ARRAY_NS(ptr_t) ]
 -# ARRAY_ptr_ptr_t [ ARRAY_NS(ptr_ptr_t) ]\n
	A pointer to an array pointer - useful in pointer allocators.
 -# ARRAY_packed_ptr_t [ ARRAY_count_t ]
 -# ARRAY_packed2_ptr_t [ ARRAY_count_t ]\n
	This is the handle-to-handle-to-array type.
 -# ARRAY_data_t [ ARRAY_NS(data_t) ]\n
	The array entry type.\n
	Any pointers it contains to allocated data must also be packed.
 -# ARRAY_input_t [ ARRAY_data_t ]\n
	Used for insertion.
	It could contain all the information required to populate a data_t, but for
	complex data_t's it's quicker and easier to just supply enough data for the
	allocation to complete and then populate the data_t directly.

6. header/implementation - optional : name [ default ]\n
 -# ARRAY_prefix [ extern ]\n
	Prefixes function prototypes/definitions.\n
	A sometimes useful alternative is [ static ]

7. header/implementation - array_t members : var [ default ]\n
 -# ARRAY_count [ m_count ]
 -# ARRAY_reserve [ m_reserve ]
 -# ARRAY_increment [ m_increment ]
 -# ARRAY_items [ m_items ]

8. implementation - must-defines\n
 -# ARRAY_alloc(n) to initially allocate array storage\n
	Required if TREEDB_OMIT_ARRAY_new(_DECLARE) aren't defined.
 -# ARRAY_realloc_packed(pha, new_sz) to reallocate the array storage.
 -# ARRAY_free_packed(n) to free an array storage\n
	Required if TREEDB_OMIT_ARRAY_free(_DECLARE) aren't defined.
 -# ARRAY_unpack_pointer(u) to unpack a pointer\n
	Unpacking/packing a pointer converts to/from the internal representation of
	the pointer. Usually this internal representation is an offset from the
	start of the "heap".
 -# ARRAY_pack_pointer(p) to pack a pointer

9. implementation - optional : var [ default ]\n
 -# ARRAY_auto_data(x) [ ARRAY_data_t x ]\n
	Stack-based instances created to reverse entries.
 -# ARRAY_clear_entries(ha, begin, size) [ ARRAY_NS(clear_entries)(runtime, ha,
	begin, size) ]\n
	to free any allocated data associated with the specified range.
 -# ARRAY_free_packed(x) [ ARRAY_free(ARRAY_unpack(x)) ]
 -# ARRAY_memmove(to, from, size) [ memmove(to, from, size) ]
 -# ARRAY_assign(to, from) [ to = from ]
 -# ARRAY_assign_to_auto(to, from) [ ARRAY_assign(to, from) ]
 -# ARRAY_assign_from_auto(to, from) [ ARRAY_assign(to, from) ]
 -# ARRAY_item(pa, n) [ (pa)->ARRAY_items[n] ]
	You will need to change this if you're writing a template implementation
	that uses "char" as the underlying storage type.
 -# ARRAY_item_ptr(pa, n) (& ARRAY_item(pa, n))\n
	Used to memmove() items for inserts / removes.
 -# ARRAY_entry_size [ sizeof(ARRAY_data_t) ]
 -# ARRAY_entry_align [ __alignof(ARRAY_data_t) ]

10. implementation - pointer pack / unpack : var [ default ]\n
 -# ARRAY_unpack(x) [ ARRAY_NS(unpack)(runtime, x) ]
 -# ARRAY_pack(x) [ ARRAY_NS(pack)(runtime, x) ]
 -# ARRAY_upack2(x) [ ARRAY_NS(upack2)(runtime, x) ]\n
	Converts a ARRAY_packed2_ptr_t to an ARRAY_ptr_t

@section helpers treedb-array-helper.h et al
treedb-array-helper-defines.h - var [ default ]\n
 -# ARRAY_alloc(i) [ ARRAY_NS(impl_alloc)(runtime, i) ]
 -# ARRAY_free(n) [ ARRAY_NS(impl_free)(runtime, n) ]
 -# ARRAY_realloc_packed(ppap, count) [
	ARRAY_NS(impl_realloc_packed)(runtime, ppap, count) ]
 -# ARRAY_unpack_pointer(p) [ TREEDB_NS(impl_unpack_pointer)(runtime, p) ]
 -# ARRAY_pack_pointer(p) [ TREEDB_NS(impl_pack_pointer)(runtime, p) ]

*/
