/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

/**
@mainpage treedb
@author Philip Ashmore
@version $(VERSION)

@section copying Copying

Copyright (C) 2010 Philip Ashmore

This library is distributed under LGPL. See the LICENSE.txt file for
details.

@page c-templates c-templates
"C" templates, or c-templates are include files that allow the includer to
control how the contents of the file are expanded.

The "template parameters" are macro definitions.

Each template requires a macro that defines the "namespace" of the template.

There are no namespaces in "C", so this is achieved by wrapping everything with
a name inside a namespace macro that the includer provides.

Unfortunately macro definitions cannot be wrapped in this way, but as long as
they're local to the include file that's OK.

C-templates consist of template members/components.

These can be anything from typedefs, macro definitions, function prototypes or
declarations.

Template members can be conditionally excluded(omitted) with macros.

Some members depend on other members, so omitting one member will cause the
dependent members to also be omitted.

In the case of AVL trees, avl.h defines the tree interface - its structures and
functions, while avl-impl.h defines the implementation.

The includer has to define various macros:
 -# what the name of the namespace is\n
	In the case of AVL trees, you need to define AVL_TMPL_NS(x) to transform
	the name "X" into something unique, like MyTreeX32
 -# specify what and how structures are defined, including the member names
 -# specify what functions are defined

The same goes for the c-template implementation(impl) include file.

With c-templates it's possible to define arrays of trees, trees of lists, lists
of arrays, etc, and even use several types defined/implemented by the same
c-template include files.

See varray.h and varray-impl.h - they use two types of arrays, and yes, they're
also c-template files.

Dont forget the *_prefix macros - if you define them to "static" then the entire
interface/implementation will be confined to the compilation unit that includes
it.

@page intro Introduction

A treedb is an allocator on top of memory obtained from mmap(), but one can be
created on top of regular memory too.

For persistent treedbs, the mmapped area is a window onto one physical file.

A treedb is a persistent memory heap that can be accessed simultaneously by
multiple threads in multiple processes.

treedb's have several runtime configuration options, the most important of which
is the heap allocator algorithm.

Currently the choices for the allocator algorithm is a two-tree allocator
where the tree implementation is one of
 -# avl
 -# red-black(TODO)
 -# others welcome!
where the "used" tree sorts nodes by address (without duplicates) and the "free"
tree sorts nodes by size, allowing duplicates.

Other allocator implementations are welcome!

@section implementation Implementation

treedb implementation has the following attributes:
 -# memory is allocated on top of a memory map (mmap) that may or may not be
	on top of a file.
 -# to allow different applications to simultaneously access/update a treedb,
	it's likely that these applications will mmap the treedb onto different
	memory addresses.\n
	To address this issue, inter-structure references (pointers) need a
	different stored representation - as offsets, referred to as
	"packed pointers".\n
 -# for offset allocators, a packed pointer is the same as an offset, and
	packing a pointer returns it's offset from the start of (treedb) storage
 -# for pointer allocators, a packed pointer is the same as a pointer, and
	packing a pointer just returns the supplied pointer
 -# there are some situations where you really need to convert to and from an
	offset, regardless of the allocator implementation, in which case
	pointer-to-offset and offset-to-pointer operations are needed
 -# the special offset "0" means "void" and packing, unpacking, pointer-to-offset
	and offset-to-pointer operations all need to take this into account
 -# to allow different applications to update via transactions, locking
	semantics are needed
 -# treedb's are platform independent, meaning that one treedb file may be
	copied to machines running different operating systems on different
	processors with different native byte and/or word ordering.\n
	To do this, treedb files store meta-data describing how their structures are
	arranged in the header.\n
 -# treedb's data should be platform independent\n
	Applications are free to store their data in any format of their choosing
	but this format should be either platform independent, or should include
	some meta-data describing how this data is represented, just like treedb.\n
	Applications don't have to adhere to this ideal for various reasons, like
	if they won't be copied between different platforms.
 -# choice of tree implementations
	- avl tree
	- red-black tree(TODO)
	- more as time allows

	This is a performance-related constraint - pick the best for your
	application.
 -# choice of array implementations
	- array\n
		This is an array whose meta-data (count, reserve, increment) is followed
		by a packed pointer or handle to the array data.\n
		The array metadata doesn't move around, as its size is fixed, but as the
		storage itself may move when resized, the array is accessed through a
		packed pointer.\n
		To get at the data, you unpack the handle to get at the array metadata
		and then unpack the items handle to get at the data.\n
		Pros\n
		 - you don't need to allocate (new) the array itself, but it does need
			to be initialised.\n
			This makes it convenient as a treedb payload member.
		Cons\n
		 - making it optional is a bit more awkward
	- handle-array\n
		This is an array whose meta-data (count, reserve, increment) is followed
		by the array data.\n
		A packed pointer to a packed pointer to the array (2!) is required as
		the array can move around as it is resized.\n
		To get at the data, you unpack the handle twice and you get the array
		metadata and the data.\n
		Pros\n
		 - you can make an array "optional" by just not allocating it and
			setting its handle-handle (hha) to 0
		Cons\n
		 - you need to allocate (new) the array itself, making it a bit more
			awkward to use as a treedb payload member

@section pathfinder Pathfinder

@subsection stages Pathfinder stages
I chose the names of galaxies as milestones for ease of reference.\n
To this end I use the names of galaxies that are well known in and to keep the
number of stages small.\n

@subsubsection andromeda Andromeda(1)
Andromeda's goal is to transform the avl tree implementation minimally, so as to
enable the original avl tree tests to become treedb tests with minimal
modification.

This is the "augmented reference" implementation, allowing any additional tests
that were written for the original avl tree implementation to be transformed
into treedb tests with little effort.

Andromeda achieves these goals as follows.
 -# add pointer/offset macros to the avl tree implementation
 -# define the storage interface (without locking)
 -# implement a storage interface that is consistent with the original code
 -# test cases are the avl tree tests
 -# write pointer and offset variants of the tests
 -# Just use the tree lock\n
    This serialises access for ease of testing/debugging.
 -# Modularise the avl tree tests so that they can compete for access to the
    same tree from multiple threads

@subsubsection cartwheel Cartwheel(2)
Cartwheel's goals are to
 -# implement a memory allocator using avl trees\n
    To test this in isolation, the "heap" is supplied from regular stack memory
    and is fixed.
 -# Use this allocator as part of a container implementation\n
    This allows the avl tests to run "inside" an avl tree allocator.
 -# test cases

@subsubsection comet Comet(3)
Comet's goals are to
 -# bootstrap a mmap allocator
    This stage is vital - treedb manages memory stored in a mmap, but it also
    needs to be stored inside the heap it manages.\n
 -# Manage heap resizing
    treedb needs to be able to at least try to resize it's heap in response to
    changes in the amount of storage it's clients require.
 -# test cases

@subsubsection sombrero Sombrero(4)
Sombrero's goals are to
 -# Manage a collection of mmapped heaps
	This allows it to become
	 - a general purpose allocator\n
		this has the same interface as cartwheel, but with the ability to add
		more heaps as needed.
 -# Manage more than one type of mmapped heap\n
	Multiple 16 bit heaps are limited to 64K in size and offer a fast and\n
	space-efficient way to store small allocations.
	Larger allocations go into 32 bit heaps, if needed, and yet larger ones\n
	go into their own heaps.\n
	These heaps can't resize, otherwise they might need to move.\n
	Fortunately, you can mmap() 64K of address space without requiring it to
	be backed by 64K of physical memory unless needed, so these are cheap.\n
	This allows sombrero to implement IMalloc for an application, with multiple
	instances if required.

@subsection summary Summary
To achieve these goals, treedb tree and node offsets won't, in general, be the
same size as offsets used to store pointers, which will be 32 or 64 bits
depending on the pointer size.

The heap manager translates from a pointer to the heap containing it, which
along with the heap's offset, gives the ultimate stored offset for the pointer.

The same works in reverse - given an offset, the heap manager, knowing the
heap for that offset and that heap's starting address in memory, can return
a pointer that refers to the data.

The great thing about implementing sombrero this way is that cartwheel's tests
can be re-used with just a few declaration changes.\n

@section cmds Commands I've run more than 100 times or are easy to forget

kdbg v3c/3-comet/tests/cpp_resizing_strings_test_16

make -C build/v3c/3-comet/tests cpp_resizing_strings_test_16

valgrind -v --leak-check=full --show-reachable=yes --db-attach=yes --db-command="kdbg -p %p %f" build/v3c/3-comet/int-tree-test-16 2>&1 | tee make.txt

@section todo to do
@subsection heap-snapshots Heap snapshots
Also known as "live backup", this feature requires that the pages are backed by
an LVM2 volume or fuse filesystem.\n
Ideally the entire partition would be a treedb.

LVM2 provides a feature called "snapshotting" whereby an instanteanuous copy of
an LVM2 partition can be made available while the "live" version continues to be
available.

To do this, the treedb would need to be temporarily write-locked while the
snapshot takes place.

Once the snapshot is available, the backup/copy can proceed at it's own pace.

LVM2 partitions also offer features that merge well with treedb, such as
resizing.

Of course, fuse can do this too.

@example 1-andromeda/tests/pointer-int-tree.c
A basic pointer avl-tree containing integers.

@example 1-andromeda/tests/pointer-string-tree.c
A basic pointer avl-tree containing strings.

@example 1-andromeda/tests/offset-fixed-int-tree-c.c
A basic offset avl-tree containing integers.

@example 1-andromeda/tests/offset-fixed-int-tree-cxx.cpp
A basic offset avl-tree containing integers - the c++ version.

@example 1-andromeda/tests/offset-fixed-string-tree-c.c
A basic offset avl-tree containing strings.

@example 1-andromeda/tests/offset-fixed-string-tree-cxx.cpp
A basic offset avl-tree containing strings - the c++ version.

@example 1-andromeda/tests/offset-fixed-int-tree-mt.c
A basic offset avl-tree containing integers - the multi-threaded version.

@example 1-andromeda/tests/offset-fixed-int-tree-mt2.c
A basic offset avl-tree containing integers - another multi-threaded version.

@example 2-cartwheel/tests/allocator-test.c
An attempt at catching all those allocator corner cases.

@example 2-cartwheel/tests/string-allocator.cpp
A treedb offset avl-tree containing strings.

The c++ build of this file shows the allocator nodes with the string data.

@example 2-cartwheel/tests/two-trees.cpp
A treedb containing two offset avl-trees, one containing strings, the other
containing integers.

The c++ build of this file shows the allocator nodes with the integer and string
data.

@example 3-comet/tests/varray-test1.cpp
A treedb varray test.

@example 3-comet/tests/varray-test2.cpp
Another treedb varray test.

@example 3-comet/tests/fork-test.c
This example is intended to create a lot of "churn" where several processes
read/write a treedb.

It is built as several variants to show the kind of debugging output you can
expect if you define TREEDB_DEBUG and/or if you link with the debugging
version of libtreedb.

The variable "fork_length" defines the length of a fork-chain of processes with
each child process having a larger process_id than its parent, starting at zero.

So, for a fork_length of 7 we get the following chain of processes with the
given process_ids.
\code
	0 -> 1 -> 2 -> 3 -> 4 -> 5 -> 6
\endcode

The "root" process creates the int treedb "fork-test.3db".

The process-sharing mechanism treedb uses with treedb files is by read or write
locking the first byte when read or write access is required, respectively.

In this test, root write-locks the second byte.

Non-root processes try to read-lock the second byte and block.

When all the child processes are started the root process waits a bit for all
processes to reach the starting line and then unlocks the second byte.

This causes an avalanche of attempts to add and remove different numbers from
the int tree, maximizing the chances of processes blocking each other.

Here are the details, also present in the code:
1. The "root" process(process_id == 0) creates an int treedb fork-test.3db.\n
	It stores the size of the only free node for later checking.

2. Root write locks the second byte.

3. Processes spawn child processes before opening the treedb.\n
	The "leaf" process (process_id == (fork_length-1)) doesn't create a child
	process.\n
	Child processes close the treedb "second byte" file handle - the write lock
	is not inherited.

4. Non-root processes block trying to read-lock the second byte.

5. Root waits a bit and unlocks the second byte.

6. Non-root processes insert their own process_id

7. Parent processes loop waiting for their childs process_id to be inserted.

8. Parent processes remove their childs process_id.

9. Non-root processes allocate a chunk of storage.

10. Non-leaf processes wait for their child to exit.

11. Non-root processes then free their chunk and exit.

12. The root process checks that the free treedb node is the same size as it was
	when the treedb was created, as a consistency check.

13. The root process then deletes fork-test.3db.

@example 3-comet/tests/int-tree-test.cpp
A treedb containing an avl-tree payload containing integers.

If built with -DTREEDB_DEBUG you get tree dumps.

It also tries to be a thorough test of the avl functions remove_many(),
remove_between(), distance() and node_index() indirectly.
@example 3-comet/tests/string-key-tree-test.cpp
A treedb with an avl-tree of structures keyed by string.

The complication with string keys is that clone() can indirectly construct avl
tree nodes, so a check is needed to see where the key "comes from".

If it's supplied by the client code then easy peasy.

If it's supplied by clone() then the treedb may move because of a heap
enlargement requiring a munmap()+mmap().

When strcpy() is called to store the string, kaboom!

The solution is to call TREEDB_in_heap(key) and store its offset before the
allocation + restore it afterwards.

This approach is needed anytime the data to store isn't supplied by value, i.e.
a pointer.

@example 3-comet/tests/string-value-tree-test.cpp
A treedb with an avl-tree of variable-sized-structures keyed by integer.

@example 3-comet/tests/struct-list-test.cpp
A treedb with a doubly-linked-list of structures.

@example 3-comet/tests/handle-array-test.cpp
A treedb with a handle-array of integers.

@example 3-comet/tests/array-test.cpp
A treedb with an array of integers.

@example 3-comet/tests/int-tree-mt-test.cpp
A treedb with an avl-tree of integers accessed by multiple threads.

@example 3-comet/tests/wordset-test.cpp
A treedb stress test.

@example 3-comet/tests/wordset-test-2.cpp
A treedb stress test.

@example 3-comet/tests/resizing-strings-test.cpp
A treedb stress test.

@example 3-comet/tests/multi-varray-test.cpp
TODO

@example 4-sombrero/tests/sombrero-test.c
A demo using an allocator whose implementation uses other allocators for
different sized requests.

@section treedb_init_functions treedb init functions
Keeping track of these is hard enough so I've documented them here.

@subsection treedb_init_size
 - set size

@subsection treedb_init_header
 - check parameters
 - assign header variables
 - treedb_init_size

@subsection treedb_init_memory
 - setup used + free trees
 - create + insert one free node

@subsection treedb_init
 - treedb_init_header
 - treedb_init_memory

@section treedb_heap_init_functions treedb heap init functions
Keeping track of these is hard enough so I've documented them here.

@subsection heap_get_size
 - get size using seek and the offset

@subsection heap_just_open_fd
 - file_lock_state_init
 - set_mmap_flags
 - treedb_init_size
 - heap_?_lock

@subsection heap_just_open
 - heap_just_open_fd(open(name))

@subsection heap_open
 - heap_just_open
 - heap_header_check

@subsection heap_open_fd
 - heap_just_open_fd
 - heap_header_check

@subsection heap_init_size
 - ftruncate
 - treedb_init_size

@subsection heap_init_fd
 - close the heap if it's already open
 - file_lock_state_init
 - set_mmap_flags
 - heap_init_size
 - heap_write_lock

@subsection heap_init
 - heap_init_fd(open(name))

@subsection heap_create
 - heap_init
 - treedb_init

*/
