/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

/*! @page multi-varray multi-varrays
A multi-varray (multi-varray) implements the varray concept.

It is implented using an avl-tree, whose nodes contain varrays.

This allows elements to be added and removed without the overhead incurred in
reallocating large blocks of memory.

It has a max_size attribute that limits the size of any varray and also the size
of an entry.

The implementation prevents any individual varray from growing too large or too
small after insertions and removals
 - adjacent varrays are joined if their contents would fit in one varray
 - varrays are split if an insertion would make a varray too large

Even though varray::clear() can fail multi-varray::clear() cannot
as it clears the tree containing the varray nodes.

Because entries are of varying size, the reserve and increment values supplied
to init() are in bytes, and are passed to the payload arrays init() function.

The average size of entries is supplied as another parameter, and is used along
with the increment value to calculate the offset array increment value.

@section setup What to do before including header / implementation
Make sure your definitions before including the header are consistent with those
before including implementation.

1. header - omit type declarations
 -# MVARRAY_omit_array_t
 -# MVARRAY_omit_array_t_typedefs

2. header/implementation - omit function prototypes/definitions
	respectively with
 -# MVARRAY_omit_new_DECLARE
 -# MVARRAY_omit_free_DECLARE

3. implementation - omit function definitions
 -# MVARRAY_omit_new
 -# MVARRAY_omit_free

4. header - must-defines\n
 -# MVARRAY_NS(x)
	Every type and function is wrapped using this macro to allow for multiple
	array "types", e.g. float arrays and int arrays.

5. header/implementation - types : type [ default ]\n
 -# MVARRAY_runtime_ptr_t [ MVARRAY_NS(runtime_ptr_t) ]\n
	This is the type of the array and is always named "runtime".\n
	This way, the internal macros don't need to be passed this as a parameter.
 -# MVARRAY_size_t [ uintptr_t ]\n
	The type used to exchange scalar information with the client.
 -# MVARRAY_count_t [ MVARRAY_size_t ]\n
	The type used in the implementation for scalars and packed pointers.
 -# MVARRAY_ptr_t [ MVARRAY_NS(ptr_t) ]
 -# MVARRAY_packed_ptr_t [ MVARRAY_count_t ]
 -# MVARRAY_data_t [ MVARRAY_NS(data_t) ]\n
	The array entry type.
	Any "pointers" it contains to allocated data must be packed.
 -# MVARRAY_input_t [ MVARRAY_data_t ]\n
	Used for insertion.
	It could contain all the information required to populate a data_t, but for
	complex data_t's it's quicker and easier to just supply enough data for the
	allocation to complete and then populate the data_t directly.

6. header/implementation - optional : name [ default ]\n
 -# MVARRAY_prefix [ extern ]\n
	Prefixes function prototypes/definitions.\n
	A sometimes useful alternative is [ static ]

7. header - constants\n
 -# VARRAY_init\n
	These defaults are completely off the top of my head :)\n
	{ 0, 0, 128 * sizeof(MVARRAY_count_t), 32 * sizeof(MVARRAY_count_t) \n
	, 16 * sizeof(MVARRAY_count_t), 8 * sizeof(MVARRAY_count_t) \n
	, 4 * sizeof(MVARRAY_count_t), AVL_TMPL_init }

7. header/implementation : array_t member names : name [ default ]\n
 -# MVARRAY_count [ m_count ]
 -# MVARRAY_length [ m_length ]
 -# MVARRAY_max_size [ m_max_size ]
 -# MVARRAY_payload_reserve [ m_payload_reserve ]
 -# MVARRAY_payload_increment [ m_payload_increment ]
 -# MVARRAY_payload_average [ m_payload_average ]
 -# MVARRAY_offset_increment [ m_offset_increment ]
 -# MVARRAY_root [ m_root ]

8. implementation - must-defines\n
 -# MVARRAY_alloc(n) to initially allocate array storage\n
	Required if MVARRAY_omit_new(_DECLARE) aren't defined.
 -# MVARRAY_realloc_packed(pha, new_sz) to reallocate the array storage.
 -# MVARRAY_free_packed(n) to free an array storage\n
	Required if MVARRAY_omit_free(_DECLARE) aren't defined.
 -# MVARRAY_unpack_pointer(u) to unpack a pointer\n
	Unpacking/packing a pointer converts to/from the internal representation of
	the pointer. Usually this internal representation is an offset from the
	start of the "heap".
 -# MVARRAY_pack_pointer(p) to pack a pointer

9. implementation - optional : var [ default ]\n
 -# MVARRAY_alloc_array() [ (invokes ARRAY_ALLOC) ]\n
	to allocate an array.
 -# MVARRAY_auto_data(x) [ MVARRAY_data_t x ]\n
	Stack-based instances created to reverse entries.
 -# MVARRAY_clear_entries(ha, begin, size) [ ARRAY_NS(clear_entries)(runtime, ha,
	begin, size) ]\n
	to free any allocated data associated with the specified range.
 -# MVARRAY_free_packed(x) [ MVARRAY_free(ARRAY_unpack(x)) ]
 -# MVARRAY_memmove(to, from, size) [ memmove(to, from, size) ]
 -# MVARRAY_assign(to, from) [ to = from ]
 -# MVARRAY_assign_to_auto(to, from) [ MVARRAY_assign(to, from) ]
 -# MVARRAY_assign_from_auto(to, from) [ MVARRAY_assign(to, from) ]
 -# MVARRAY_item(pa, n) [ (pa)->MVARRAY_items[n] ]
	You will need to change this if you're writing a template implementation
	that uses "char" as the underlying storage type.
 -# MVARRAY_item_ptr(pa, n) (& MVARRAY_item(pa, n))\n
	Used to memmove() items for inserts / removes.
 -# MVARRAY_entry_size [ sizeof(MVARRAY_data_t) ]
 -# MVARRAY_entry_align [ __alignof(MVARRAY_data_t) ]

10. implementation - pointer pack / unpack : var [ default ]\n
 -# MVARRAY_unpack(x) [ MVARRAY_NS(unpack)(runtime, x) ]
 -# MVARRAY_pack(x) [ MVARRAY_NS(pack)(runtime, x) ]

@section helpers treedb-array-helper.h et al
treedb-array-helper-defines.h : var [ default ]\n
 -# MVARRAY_alloc(i) [ MVARRAY_NS(impl_alloc)(runtime, i) ]
 -# MVARRAY_free(n) [ MVARRAY_NS(impl_free)(runtime, n) ]
 -# MVARRAY_realloc_packed(ppap, count) [
	MVARRAY_NS(impl_realloc_packed)(runtime, ppap, count) ]
 -# MVARRAY_unpack_pointer(p) [ TREEDB_NS(impl_unpack_pointer)(runtime, p) ]
 -# MVARRAY_pack_pointer(p) [ TREEDB_NS(impl_pack_pointer)(runtime, p) ]

*/
