/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

/**
@page treedb-malloc treedb's malloc demo
This demo uses treedb's allocator as a stand-in for libc's malloc.

The idea is to allow you to run memory intensive tasks without eating into swap
space, by using the operating systems caching memory like a rug, caching as much
of the treedb file-storages as it can.

The more cache memory you have available, the less you'll notice the difference.

Where you will notice is when a program would start to require swap memory.

Because treedb's malloc implementation uses memory maps exclusively, there is
no impact on available system memory, so instead the cache memory is thrown
around, like a rug that's too small trying to cover what it can.

The result is a lot of disk activity instead of your system grinding to a halt.

Memory used by applications to cache on-disk resources still gets paged out,
but not nearly as much.

@section how-it-works how it works
A program linked with the treedb malloc library or with that library preloaded
with LD_PRELOAD, creates a file as a storage for a treedb allocator.

This happens when the process starts.

This file is located in a fuse filesystem.

This is to allow the memory mapped file to have a fixed maximum size, while the
treedb-malloc-daemon resizes it only if it needs to.

So demand for available memory is converted into a demand for disk space.

@subsection why-fuse why fuse?
Processes can fork, and when they do, the parent and child share memory.

Each can write to their memory and only then are private chunks created to hold
the changes each makes, without affecting each other.

This is a lot like a quilt with holes in it. The holes show through the memory
as it existed before the fork, and the filled bits are what each processes
changed, visible only to themselves, and their children.

That's right, this chain of parent child links can be of any length, with
processes sharing memory they haven't changed with each other.

The fuse file system implemented by the daemon does this without the client
processes needing to know.

@section lib-treedb-malloc treedb malloc library
This library implements the malloc functions.

You can link your application with it and as long as there's a
treedb-malloc-daemon available, it will implement malloc in your program.

More interestingly, you can also LD_PRELOAD this library, and existing
applications will use treedbs malloc implentation - great for demos.

@subsection environment-vars environment variables
This library needs the environment variable V3C_TREEDB_MALLOC_FUSE_DIR
to contain the name of the fuse directory where the treedb-malloc-daemon is set
up to process requests.

The directory specified by V3C_TREEDB_MALLOC_FUSE_DIR has to be a fuse
filesystem, managed by treedb malloc daemon, because the daemon implements
memory sharing between forked processes.

@subsection safety-net safety net
As only a treedb-malloc-daemon fuse file system will do, it implements a
write-only control file to create overlays from mappings for a parent process,
and allows processes to open files that aren't there, a sort of signature,
assuring the treedb malloc library that it's the right kind of directory.

The other gotcha is that a non-fuse directory/filesystem would try to honour
a request to create a file capable of storing any programs memory requirements.
\code
#define KIBIBYTE ((size_t)1024)
#define MEBIBYTE (KIBIBYTE * KIBIBYTE)
#define GIBIBYTE (KIBIBYTE * MEBIBYTE)
\endcode
On a 32 bit machine that's a gibibyte, as we're address-space limited, and on a
64 bit machine that's 256 gibibytes!

A process connects to the daemon as follows

1. treedb_malloc_init - a processes is exec()'d with the library loaded
   The process may have been already connected to the daemon through a fork.
2. atfork_after_child
   The new process declares itself as being the child of an already connected
   parent.

parent backing creation
=======================
1. remove pid and temp files if the pid is present in g_runtimes
2. create backing temp files
3. insert into g_backings, keyed by pid
4. insert into g_runtimes, keyed by pid

child overlay creation
======================
1. insert tuple into g_forks
   if succeeded:
	2a. anonymize the parent runtime
	2b. overlay parent = anonymized parent runtime
   else: parent got here first after the fork
	2c. overlay parent = parents runtime parent (the anonymized runtime)
	2d. delete from g_forks
2. create overlay temp files
3. insert into g_overlays, keyed by pid
4. insert into g_runtimes, keyed by pid

parent calls "overlay my memory map"
====================================
1. insert tuple(ppid, 0) into g_forks as we don't know the childs pid yet
   if succeeded:
	2a. anonymize our runtime
	2b. overlay parent = anonymized runtime
   else: child got here first after the fork
	2c. overlay parent = our runtime
	2d. delete from g_forks
3. create overlay temp files
4. insert into g_overlays, keyed by ppid
5. insert into g_runtimes, keyed by ppid

@section treedb-malloc-daemon treedb malloc daemon
The treedb-malloc library (or library for short) connects to the
treedb-malloc-daemon (or daemon for short) when a process starts.

To use the daemon, the library creates a "/pid" file, wheere pid is it's process
id.

In the event of a fork, two things happen, in an unspecified order:
1. the parent sends an "overlay my mapping" control message
2. the child opens a file named "/ppid-pid"

In either case we
1. add the (p)pid to the g_forks set if it's not there, remove it otherwise
2. anonymize the mapping if the (p)pid was inserted into the g_forks set
3. create an overlay for the child process

Once both the control-file-write and the childs open-file happen, both
processes have overlays that store changes to their shared mapping.

This shared mapping isn't something that either process can know about or
affect, so it's called an anonymized mapping.

Because it started out as belonging to the parent, and since the parent is now
using an overlay with the same name, the anonymized mapping can't be listed.

When a process completes, its overlay is deleted, leaving one overlay onto the
anonymized mapping file.

This overlay isn't strictly speaking necessary anymore as there's only one
client process, so the overlaid changes could be merged back into the anonymized
mapping, and then de-anonymizing the backing to be owned by that process, as if
the fork never happened.

For a large number of overlaid pages, however, this could involve a significant
amount of copying between files - the process would appear to stall.

Leaving it as is means less delay for clients at the expense of some processes
having more mapping files (one backing and one or more overlays) than they
strictly need.
*/

