/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

/*! @page varray variable-arrays
A varray implements the array concept.

A varray stores items of varying length by packing them into a "payload" array
given an alignment constraint and storing their offsets in an "offset" array.

This implementation uses treedb-arrays.

For every n elements the payload array stores, the offset array needs to store
n+1 offsets to track the end of the last item.

This simplifies the implementation.

The offset array is initialised with a "0" entry denoting the offset of the end
of storage in the data array.

It is for this reason that clear() can fail - it needs to allocate one offset
entry.

Because entries are of varying size, the reserve and increment values supplied
to init() are in bytes, and are passed to the payload arrays init() function.

The average size of entries is supplied as another parameter, and is used along
with the increment value to calculate the offset array increment value.

@section setup What to do before including header / implementation
Make sure your definitions before including the header are consistent with those
before including implementation.

1. header - omit type declarations
 -# VARRAY_omit_array_t
 -# VARRAY_omit_array_t_typedefs

2. header+implementation - omit function declarations+definitions
 -# VARRAY_omit_init_empty_DECLARE
 -# VARRAY_omit_new_DECLARE
 -# VARRAY_omit_free_DECLARE
 -# VARRAY_omit_insert_DECLARE
 -# VARRAY_omit_remove_DECLARE
 -# VARRAY_omit_push_back_DECLARE
 -# VARRAY_omit_push_back_many_DECLARE
 -# VARRAY_omit_push_front_DECLARE
 -# VARRAY_omit_push_front_many_DECLARE
 -# VARRAY_omit_pop_back_DECLARE
 -# VARRAY_omit_pop_back_many_DECLARE
 -# VARRAY_omit_pop_front_DECLARE
 -# VARRAY_omit_pop_front_many_DECLARE
 -# VARRAY_omit_reverse_DECLARE

3. implementation - omit member definitions
 -# VARRAY_omit_array_t_member_defines

4. implementation - omit function definitions
 -# VARRAY_omit_init_empty
 -# VARRAY_omit_new
 -# VARRAY_omit_free
 -# VARRAY_omit_insert
 -# VARRAY_omit_remove
 -# VARRAY_omit_push_back
 -# VARRAY_omit_push_back_many
 -# VARRAY_omit_push_front
 -# VARRAY_omit_push_front_many
 -# VARRAY_omit_pop_back
 -# VARRAY_omit_pop_back_many
 -# VARRAY_omit_pop_front
 -# VARRAY_omit_pop_front_many
 -# VARRAY_omit_reverse

5. header - must-defines\n
 -# VARRAY_NS(x)
	Every type and function is wrapped using this macro to allow for multiple
	varray "types".

6. header/implementation - types : type [ default ]\n
 -# VARRAY_runtime_ptr_t [ ARRAY_NS(runtime_ptr_t) ]\n
	This is the type of the array and is always named "runtime".\n
	This way, the internal macros don't need to be passed this as a parameter.
 -# VARRAY_size_t\n
	The type used to exchange scalar information with the client.
 -# VARRAY_entry_size_t [ VARRAY_size_t ]\n
	The type used to exchange entry size information with the client.
 -# VARRAY_count_t [ VARRAY_size_t ]\n
	The type used in the implementation for scalars and packed pointers.
 -# VARRAY_ptr_t [ ARRAY_NS(ptr_t) ]
 -# VARRAY_packed_ptr_t [ ARRAY_count_t ]
 -# VARRAY_data_t [ ARRAY_NS(data_t) ]\n
	The array entry type.
	Any "pointers" it contains to allocated data must be packed.
 -# VARRAY_input_t [ VARRAY_data_t ]\n
	Used for insertion.
	It could contain all the information required to populate a data_t, but for
	complex data_t's it's quicker and easier to just supply enough data for the
	allocation to complete and then populate the data_t directly.

7. header - constants\n
 -# VARRAY_init
 -# VARRAY_count_limit

8. header/implementation - optional : name [ default ]\n
 -# VARRAY_prefix [ extern ]\n
	Prefixes function prototypes/definitions.\n
	A sometimes useful alternative is [ static ]

9. header/implementation : array_t members : var [ default ]\n
 -# VARRAY_offset [ m_offset ]
 -# VARRAY_data [ m_data ]

10. implementation - must-defines\n
 -# VARRAY_alloc(n) to initially allocate array storage\n
	Required if VARRAY_omit_new(_DECLARE) aren't defined.
 -# VARRAY_realloc_packed(pha, new_sz) to reallocate the array storage.
 -# VARRAY_free_packed(n) to free an array storage\n
	Required if VARRAY_omit_free(_DECLARE) aren't defined.
 -# VARRAY_unpack_pointer(u) to unpack a pointer\n
	Unpacking/packing a pointer converts to/from the internal representation of
	the pointer. Usually this internal representation is an offset from the
	start of the "heap".
 -# VARRAY_pack_pointer(p) to pack a pointer

11. implementation - optional : var [ default ]\n
 -# VARRAY_alloc_array() [ (invokes ARRAY_ALLOC) ]\n
	to allocate an array.
 -# VARRAY_auto_data(x) [ ARRAY_data_t x ]\n
	Stack-based instances created to reverse entries.
 -# VARRAY_clear_entries(ha, begin, size) [ ARRAY_NS(clear_entries)(runtime, ha,
	begin, size) ]\n
	to free any allocated data associated with the specified range.
 -# VARRAY_free_packed(x) [ ARRAY_free(ARRAY_unpack(x)) ]
 -# VARRAY_memmove(to, from, size) [ memmove(to, from, size) ]
 -# VARRAY_assign(to, from) [ to = from ]
 -# VARRAY_assign_to_auto(to, from) [ ARRAY_assign(to, from) ]
 -# VARRAY_assign_from_auto(to, from) [ ARRAY_assign(to, from) ]
 -# VARRAY_item(pa, n) [ (pa)->VARRAY_items[n] ]
	You will need to change this if you're writing a template implementation
	that uses "char" as the underlying storage type.
 -# VARRAY_item_ptr(pa, n) (& ARRAY_item(pa, n))\n
	Used to memmove() items for inserts / removes.
 -# VARRAY_entry_align [ TREEDB_ALIGN ]

12. implementation - pointer pack / unpack : var [ default ]\n
 -# VARRAY_unpack(x) [ ARRAY_NS(unpack)(runtime, x) ]
 -# VARRAY_pack(x) [ ARRAY_NS(pack)(runtime, x) ]

@section helpers treedb-array-helper.h et al
treedb-array-helper-defines.h : var [ default ]\n
 -# VARRAY_alloc(i) [ ARRAY_NS(impl_alloc)(runtime, i) ]
 -# VARRAY_free(n) [ ARRAY_NS(impl_free)(runtime, n) ]
 -# VARRAY_realloc_packed(ppap, count) [
	VARRAY_NS(impl_realloc_packed)(runtime, ppap, count) ]
 -# VARRAY_unpack_pointer(p) [ TREEDB_NS(impl_unpack_pointer)(runtime, p) ]
 -# VARRAY_pack_pointer(p) [ TREEDB_NS(impl_pack_pointer)(runtime, p) ]
*/
