/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#ifndef treedb_treedb_header_h_included
#define treedb_treedb_header_h_included

#include <stdint.h>
#include <uuid.h>

#ifdef __cplusplus
extern "C" {
#endif
/*!
A TreedbHeader is the "entry point" to a file-based mmap.
*/
typedef struct TreedbHeader
{
	char m_szMagic[4]; /* "3db" */
	/*! The allocator algorithm GUID

	The allocator algorithm manages the heap storage, allowing the client to
	add and remove data from the treedb "payload" dynamically.

	If the treedb is only accessed from memory, a pointer (or more usually an
	offset) in a program variable allows access to that data. Remember, resizing
	the heap may result in the heap changing it's address.

	If the treedb is backed by a physical file, then these dynamic allocations
	are stored in a file persistently, and can be accessed as long as the treedb
	payload provides a route to the allocated data, such as a tree or an offset.

	You have to start from known locations in this case, that's what the payload
	is for.

	Just as with regular memory allocations, it's possible to "leak"
	on-disk-memory persistently, so the usual practice is to provide a client
	interface and make sure it doesn't leak.

	The simplest way to allow anyone to define their own allocator algorithm is
	to specify it by GUID.

	I defined two myself - currently avl is provided by treedb, but this should
	not deter anyone from implementing their own - just publish some code to
	handle them should they find their way into the wild!

	avl       : 7c681d8e-99b0-493a-95ff-2373409845c7

	red-black : d8cfb953-dbce-4854-87cf-c2bdc946544b

	*/
	uuid_t m_uAllocatorAlgorithm;
	/*! The schema GUID

	The schema defines the payload of the treedb - data entry points located
	after the header, such as tree roots and other offsets to allocated data.
	It should be independent of the offset size and flags.

	In other words, it shouldn't depend on the allocator algorithm used.

	Of course, for trees and offsets within the treedb, the size of these
	offsets should match those used by the allocator, and here's why:
	 - too small an offset size won't cover all possible offsets returned by the
		allocator
	 - too large an offset size wastes space

	So you may need to look at m_uBytesPerOffset to see it's offset size when
	accessing the payload.
	*/
	uuid_t m_uSchema;
	/*! The payload offset / header size

	Applications need to store "payload data" after the header, and the
	header size (including the allocator) depends on the algorithm and it's
	flags, so this is a quick way to get it. */
	uint64_t m_ulPayloadOffset;
	/*! Bytes per offset*/
	uint8_t m_uBytesPerOffset;
	/*! Endian index

	A '0' at bit 0 means that 0x00FF is stored in "little endian"
	format [00] [FF].

	A '1' at bit 0 means that 0x00FF is stored in "big endian"
	format [FF] [00].

	This is how bytes are combined to represent words.

	bit 1 defines how words are combined to produce double-words.

	bit 2 defines how double-words are combined to produce quad-words.

	Range : 0 - 7. */
	uint8_t m_uEndianIndex;
	/*! Algorithm-dependent flags indicating what options are configured
	for the algorithm

	For avl trees, we have
	 - bit 0: "count" node field
	 - bit 1: "depth" node field
	*/
	uint32_t m_uFeatures;
} TreedbHeader, * TreedbHeaderPtr;

// 7c681d8e-99b0-493a-95ff-2373409845c7
UUID_DEFINE
( UUID_treedb_avl_two_tree_allocator_algorithm
, 0x7c, 0x68, 0x1d, 0x8e
, 0x99, 0xb0
, 0x49, 0x3a
, 0x95, 0xff
, 0x23, 0x73, 0x40, 0x98, 0x45, 0xc7
);
// d8cfb953-dbce-4854-87cf-c2bdc946544b
UUID_DEFINE
( UUID_treedb_redblack_two_tree_allocator_algorithm
, 0xd8, 0xcf, 0xb9, 0x53
, 0xdb, 0xce
, 0x48, 0x54
, 0x87, 0xcf
, 0xc2, 0xbd, 0xc9, 0x46, 0x54, 0x4b
);

#ifdef __cplusplus
} /*extern "C"*/
#endif

#endif /* #ifndef treedb_treedb_header_h_included */
